/* $Id$
 * env_envctrl.c: Environmental monitoring and front status panel
 *                support for SUNW,envctrl (E450)
 *
 * - THIS IS A PRELIMINARY DRIVER INTENDED FOR DEVELOPMENT PURPOSES -
 *
 * NOTE: Malfunction of this driver may render your system permanently
 *       damaged.  By using this driver you assume all risks due to
 *       improper operation or otherwise.
 *
 * Copyright (C) 2004-2007 Eric Brower (ebrower@gmail.com)
 * Copyright (C) 2006-2007 David Johnson (dj@david-web.co.uk)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * TODOs: please see env_envctrltwo.c for TODOs and IDEAs 
 *        Fix fan_wdt timer duration / relationship with env_mon_interval
 *        Cleanup debugging macros
 *        Define structs for thermisters, psus, disks, etc.
 *        Distinguish between different PSU states
 *        Set maintenance LED when subsystems have an error condition
 */

#include <linux/module.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/delay.h>
#include <linux/ioport.h>
#include <linux/init.h>
#include <linux/miscdevice.h>
#include <linux/mm.h>
#include <linux/slab.h>
#include <linux/kernel.h>
#include <linux/i2c.h>
#include <linux/kmod.h>
#include <linux/interrupt.h>
#include <linux/kthread.h>
#include <linux/err.h>
#include <linux/spinlock.h>
#include <linux/reboot.h>
#include <linux/sysfs.h>
#include <linux/platform_device.h>

#include <asm/of_device.h>
#include <asm/uaccess.h>
#include <asm/auxio.h>

#include <asm/i2c-envctrl.h>
#include "env_envctrl.h"
#include "env_lib.h"

#define ENV_DEVNAME	"envctrl"
#define ENV_VERSION	"0.2.2"

/* 
 * Downrev OBP versions (< 3.12 ???) provide incomplete properties--
 * we create default values for tables, but ignore CPU 1-3 thermister
 * table entries because it is not easy to tell if they are actually present
 * and I'm not sure what values will be returned if we query them.
 *
 * TODO: test if cpu node "upa-id" values correspond to CPU IDs
 *       (is it worth the effort, or just warn the user?)
 *       or provide env_cpuign="1234" module parameter for ignoring 
 *	 CPUs (see "thermister->disabled" structure member)...
 */
#define ENV_OBP_IGN_CPUS (1) /* ignore CPUs 1-3 when OBP is downrev       */
#define ENV_ENABLE_INTR  (1) /* set to enable interrupts (experimental)   */
#define ENV_DEBUG_INTR   (1) /* set to enable interrupt debugging         */
#define ENV_DEBUG_I2CDEV (0) /* scan I2C bus for devices                  */
#define ENV_DEBUG_THERMS (0) /* dump thermisters found and values probed  */
#define ENV_DEBUG_FAN    (0) /* user fanspeed control via sysfs - DANGER! */
#define ENV_DEBUG_TEMP   (0) /* user CPU temp control via sysfs           */
#define ENV_DEBUG_RAW    (0) /* ro raw sensor data via sysfs              */

static struct pcf_envctrl_nexus *nex;
static struct i2c_adapter *envctrl_adap;

/*
 * OBP "thermisters" property
 * Note: "name" is actually variable length,
 * but null-padded to at least 4 characters
 */
struct env_thermister_prop {
	u32	addr;
	u32	port;
	u32	min_temp;
	u32	warn_temp;
	u32	shutdown_temp;
	u32	num;
	u32	den;
	u8	name[4];
}__attribute((packed));

typedef enum env_therm_type {
	therm_type_unk = 0,
	therm_type_cpu,
	therm_type_noncpu,
	therm_type_lm75
} thermtype_t;

struct therm {
	struct env_thermister_prop  *cfg;
	thermtype_t                  type; 
	unsigned long                last_warn;
	unsigned int	             disabled;
	struct list_head             list;
};

struct fan {
	u8 fanport;
	u8 last_speed;
	bool forced;
};

static struct fan env_cpu_fans = {
	.fanport = ENV_FAN_PORT_CPU,
	.last_speed = 0,
	.forced = 0,
};

static struct fan env_psu_fans = {
	.fanport = ENV_FAN_PORT_PS,
	.last_speed = 0,
	.forced = 0,
};

static struct fan env_afb_fan = {
	.fanport = ENV_FAN_PORT_AFB,
	.last_speed = 0,
	.forced = 0,
};

struct threefan_status {
	bool fan0state;
	bool fan1state;
	bool fan2state;
};

struct onefan_status {
	bool fanstate;
};

static struct threefan_status env_cpu_fans_state = {
	.fan0state = 0,
	.fan1state = 0,
	.fan2state = 0,
};

static struct threefan_status env_psu_fans_state = {
	.fan0state = 0,
	.fan1state = 0,
	.fan2state = 0,
};

/*static struct onefan_status env_afb_fan_state = {
	.fanstate = 1,
};*/

spinlock_t fan_states_lock;

static struct list_head env_therms = LIST_HEAD_INIT(env_therms);

/* default table mapping CPU thermister reading to temperature Celsius,
 * indexed by raw thermister reading.  This is overridden if the OBP
 * cpu-temp-factors property exists.
 */
static u8 env_cpu_temp_table[256] = {
	0x96, 0x96, 0x96, 0x96, 0x96, 0x96, 0x96, 0x96, /*   0 -   7 */
	0x96, 0x96, 0x96, 0x96, 0x96, 0x96, 0x96, 0x96, /*   8 -  15 */
	0x96, 0x96, 0x96, 0x96, 0x96, 0x96, 0x96, 0x96, /*  16 -  23 */
	0x96, 0x96, 0x96, 0x94, 0x92, 0x90, 0x8f, 0x8e, /*    ...    */
	0x8d, 0x8c, 0x8a, 0x88, 0x87, 0x86, 0x85, 0x84, 
	0x83, 0x82, 0x81, 0x80, 0x7f, 0x7e, 0x7d, 0x7c, 
	0x7b, 0x7a, 0x79, 0x79, 0x78, 0x78, 0x77, 0x76, 
	0x75, 0x74, 0x73, 0x72, 0x71, 0x70, 0x70, 0x6f, 
	0x6f, 0x6e, 0x6e, 0x6e, 0x6d, 0x6d, 0x6c, 0x6b, 
	0x6a, 0x6a, 0x69, 0x69, 0x68, 0x67, 0x66, 0x65, 
	0x65, 0x64, 0x64, 0x64, 0x63, 0x63, 0x62, 0x62, 
	0x61, 0x61, 0x60, 0x60, 0x5f, 0x5f, 0x5e, 0x5e, 
	0x5d, 0x5d, 0x5c, 0x5c, 0x5b, 0x5b, 0x5b, 0x5a, 
	0x5a, 0x5a, 0x59, 0x59, 0x58, 0x58, 0x57, 0x57, 
	0x56, 0x56, 0x55, 0x55, 0x54, 0x54, 0x53, 0x53, 
	0x52, 0x52, 0x52, 0x51, 0x51, 0x50, 0x50, 0x50, 
	0x50, 0x4f, 0x4f, 0x4f, 0x4e, 0x4e, 0x4e, 0x4d, 
	0x4d, 0x4d, 0x4c, 0x4c, 0x4c, 0x4b, 0x4b, 0x4b, 
	0x4a, 0x4a, 0x4a, 0x49, 0x49, 0x49, 0x48, 0x48, 
	0x48, 0x47, 0x47, 0x47, 0x46, 0x46, 0x46, 0x46, 
	0x45, 0x45, 0x45, 0x44, 0x44, 0x44, 0x44, 0x43, 
	0x43, 0x43, 0x43, 0x42, 0x42, 0x42, 0x42, 0x41, 
	0x41, 0x41, 0x40, 0x40, 0x40, 0x3f, 0x3f, 0x3f, 
	0x3e, 0x3e, 0x3e, 0x3d, 0x3d, 0x3d, 0x3d, 0x3c, 
	0x3c, 0x3c, 0x3c, 0x3b, 0x3b, 0x3b, 0x3a, 0x3a, 
	0x3a, 0x39, 0x39, 0x39, 0x38, 0x38, 0x38, 0x38, 
	0x37, 0x37, 0x37, 0x37, 0x36, 0x36, 0x36, 0x35, 
	0x35, 0x35, 0x34, 0x34, 0x34, 0x33, 0x33, 0x33, 
	0x33, 0x32, 0x32, 0x32, 0x31, 0x31, 0x31, 0x30, 
	0x30, 0x30, 0x2f, 0x2f, 0x2f, 0x2e, 0x2e, 0x2e, 
	0x2d, 0x2d, 0x2d, 0x2c, 0x2c, 0x2c, 0x2b, 0x2b, 
	0x2b, 0x2a, 0x2a, 0x2a, 0x29, 0x29, 0x29, 0x28  /* 248 - 255 */
};

/* default table mapping PS thermister reading to temperature Celsius,
 * indexed by raw thermister reading.  This is overridden if the OBP
 * ps-temp-factors property exists.
 */
static u8 env_ps_temp_table[256] = {
	/* TODO: this differs from OBP ps-temp-factors on UE450... */
	0xa0, 0x9b, 0x9a, 0x96, 0x82, 0x7d, 0x78, 0x73, /*   0 -   7 */
	0x6e, 0x6e, 0x6a, 0x67, 0x65, 0x64, 0x61, 0x5e, /*   8 -  15 */
	0x5c, 0x5a, 0x58, 0x56, 0x54, 0x53, 0x52, 0x51, /*    ...    */
	0x50, 0x4f, 0x4e, 0x4d, 0x4c, 0x4a, 0x48, 0x47,
	0x46, 0x45, 0x44, 0x43, 0x42, 0x41, 0x40, 0x3f,
	0x3e, 0x3e, 0x3d, 0x3d, 0x3c, 0x3c, 0x3c, 0x3b,
	0x3b, 0x3a, 0x3a, 0x39, 0x38, 0x38, 0x37, 0x37,
	0x36, 0x36, 0x35, 0x35, 0x34, 0x34, 0x33, 0x33,
	0x32, 0x32, 0x32, 0x31, 0x31, 0x31, 0x31, 0x30,
	0x30, 0x30, 0x30, 0x2f, 0x2f, 0x2f, 0x2f, 0x2e,
	0x2e, 0x2e, 0x2d, 0x2c, 0x2b, 0x2a, 0x29, 0x29,
	0x28, 0x28, 0x28, 0x28, 0x27, 0x27, 0x27, 0x26,
	0x26, 0x26, 0x25, 0x25, 0x24, 0x24, 0x24, 0x23,
	0x23, 0x23, 0x23, 0x22, 0x22, 0x22, 0x21, 0x21,
	0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x1f, 0x1f,
	0x1f, 0x1f, 0x1f, 0x1e, 0x1e, 0x1e, 0x1d, 0x1d,
	0x1d, 0x1d, 0x1d, 0x1d, 0x1c, 0x1c, 0x1c, 0x1c,
	0x1c, 0x1c, 0x1b, 0x1b, 0x1b, 0x1b, 0x1b, 0x1a,
	0x1a, 0x1a, 0x1a, 0x1a, 0x1a, 0x1a, 0x1a, 0x1a,
	0x19, 0x19, 0x19, 0x19, 0x18, 0x18, 0x17, 0x17,
	0x16, 0x16, 0x15, 0x15, 0x15, 0x15, 0x15, 0x15,
	0x14, 0x14, 0x14, 0x14, 0x13, 0x13, 0x13, 0x13,
	0x13, 0x12, 0x12, 0x12, 0x12, 0x12, 0x11, 0x11,
	0x11, 0x11, 0x11, 0x10, 0x10, 0x10, 0x10, 0x0f,
	0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0e, 0x0e, 0x0e,
	0x0e, 0x0e, 0x0d, 0x0d, 0x0d, 0x0d, 0x0c, 0x0c,
	0x0c, 0x0c, 0x0c, 0x0b, 0x0b, 0x0b, 0x0b, 0x0b,
	0x0a, 0x0a, 0x0a, 0x0a, 0x0a, 0x0a, 0x0a, 0x0a,
	0x09, 0x09, 0x09, 0x09, 0x09, 0x09, 0x08, 0x08,
	0x08, 0x08, 0x08, 0x07, 0x07, 0x07, 0x07, 0x07,
	0x07, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06,
	0x05, 0x05, 0x05, 0x05, 0x05, 0x05, 0x05, 0x04, /* 248 - 255 */
};

#define ENV_FANTBL_SZ 112
static u8 env_cpu_fanspeed_table[ENV_FANTBL_SZ];
static u8 env_ps_fanspeed_table[ENV_FANTBL_SZ];

static u8 *env_therms_data;
static u8 env_therms_data_default[] = {
	/* CPU0 */
	0x00, 0x00, 0x00, 0x9e, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x05, 0x00, 0x00, 0x00, 0x46, 
	0x00, 0x00, 0x00, 0x50, 0x00, 0x00, 0x00, 0x01, 
	0x00, 0x00, 0x00, 0x01, 0x43, 0x50, 0x55, 0x30,
	0x00,

#if ( !ENV_OBP_IGN_CPUS )
	/* CPU1 */
	0x00, 0x00, 0x00, 0x9e, 0x00, 0x00, 0x00, 0x01,
	0x00, 0x00, 0x00, 0x05, 0x00, 0x00, 0x00, 0x46, 
	0x00, 0x00, 0x00, 0x50, 0x00, 0x00, 0x00, 0x01, 
	0x00, 0x00, 0x00, 0x01, 0x43, 0x50, 0x55, 0x31,
	0x00,

	/* CPU2 */
	0x00, 0x00, 0x00, 0x9e, 0x00, 0x00, 0x00, 0x02,
	0x00, 0x00, 0x00, 0x05, 0x00, 0x00, 0x00, 0x46, 
	0x00, 0x00, 0x00, 0x50, 0x00, 0x00, 0x00, 0x01, 
	0x00, 0x00, 0x00, 0x01, 0x43, 0x50, 0x55, 0x32,
	0x00,

	/* CPU3 */
	0x00, 0x00, 0x00, 0x9e, 0x00, 0x00, 0x00, 0x03,
	0x00, 0x00, 0x00, 0x05, 0x00, 0x00, 0x00, 0x46, 
	0x00, 0x00, 0x00, 0x50, 0x00, 0x00, 0x00, 0x01, 
	0x00, 0x00, 0x00, 0x01, 0x43, 0x50, 0x55, 0x33,
	0x00,
#endif /* ENV_OBP_IGN_CPUS */

	/* PS0 */
	0x00, 0x00, 0x00, 0x90, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x05, 0x00, 0x00, 0x00, 0x5a,
	0x00, 0x00, 0x00, 0x64, 0x00, 0x00, 0x00, 0x01,
	0x00, 0x00, 0x00, 0x01, 0x50, 0x53, 0x30, 0x00,

	/* PS1 */
	0x00, 0x00, 0x00, 0x92, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x05, 0x00, 0x00, 0x00, 0x5a, 
	0x00, 0x00, 0x00, 0x64, 0x00, 0x00, 0x00, 0x01,
	0x00, 0x00, 0x00, 0x01, 0x50, 0x53, 0x31, 0x00,

	/* PS2 */
	0x00, 0x00, 0x00, 0x94, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x05, 0x00, 0x00, 0x00, 0x5a,
	0x00, 0x00, 0x00, 0x64, 0x00, 0x00, 0x00, 0x01,
	0x00, 0x00, 0x00, 0x01, 0x50, 0x53, 0x32, 0x00, 

	/* AMB0 */
	0x00, 0x00, 0x00, 0x9a, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x05, 0x00, 0x00, 0x00, 0x28, 
	0x00, 0x00, 0x00, 0x32, 0x00, 0x00, 0x00, 0x01, 
	0x00, 0x00, 0x00, 0x01, 0x41, 0x4d, 0x42, 0x00
};

static struct timer_list env_fsp_act_timer;
static unsigned int env_fsp_irq;
static u8 env_pstate = 0xFF;

static struct env_disk_backplane {
	u32 mask;
} env_disk_bp[ENV_DISK_BP_MAX];

static struct task_struct *kenvctrld_task;

static void env_do_poweroff(void);
static void env_do_shutdown(void);
static void (*env_poweroff_pfn)(void) = env_do_shutdown;
static int env_fsp_act_timer_init = 0;

/* sysfs sh!te goes here... */
static struct platform_device *env_device;

static ssize_t env_show_fanspeed(struct device *d, struct device_attribute *attr, char *buf);
#if ENV_DEBUG_FAN
	static ssize_t env_store_fanspeed(
		struct device *d, struct device_attribute *attr, const char *buf, size_t cnt);
	static DEVICE_ATTR(fanspeed, \
		S_IRUGO|S_IWUSR, env_show_fanspeed, env_store_fanspeed);
#else 
	static DEVICE_ATTR(fanspeed, S_IRUGO, env_show_fanspeed, NULL);
#endif
#if ENV_DEBUG_TEMP
	static u8 env_forced_temp;
	static ssize_t env_store_temp(struct device *d, struct device_attribute *attr, const char *buf, size_t cnt);
	static ssize_t env_show_temp (struct device *d, struct device_attribute *attr, char *buf);
	static DEVICE_ATTR(forced_temp, \
		S_IRUGO | S_IWUSR, env_show_temp, env_store_temp);

	static ssize_t env_store_lm75(struct device *d, struct device_attribute *attr, const char *buf, size_t cnt);
	static ssize_t env_show_lm75(struct device *d, struct device_attribute *attr, char *buf);
	static DEVICE_ATTR(lm75, \
		S_IRUGO | S_IWUSR, env_show_lm75, env_store_lm75);

	static ssize_t env_store_8583(struct device *d, struct device_attribute *attr, const char *buf, size_t cnt);
	static ssize_t env_show_8583(struct device *d, struct device_attribute *attr, char *buf);
	static DEVICE_ATTR(8583, \
		S_IRUGO | S_IWUSR, env_show_8583, env_store_8583);
#endif
static ssize_t env_show_thermisters(struct device *d, struct device_attribute *attr, char *buf);
static DEVICE_ATTR(thermisters, S_IRUGO, env_show_thermisters, NULL);
/* TODO: store_power for PS status override ??? */
static ssize_t env_show_power(struct device *d, struct device_attribute *attr, char *buf);
static DEVICE_ATTR(power, S_IRUGO, env_show_power, NULL);
static ssize_t env_show_disks(struct device *d, struct device_attribute *attr, char *buf);
static ssize_t env_store_disks(struct device *d, struct device_attribute *attr, const char *buf, size_t cnt);
static DEVICE_ATTR(disks, S_IRUGO|S_IWUSR, env_show_disks, env_store_disks);
static ssize_t env_show_keyswitch(struct device *d, struct device_attribute *attr, char *buf);
static DEVICE_ATTR(keyswitch, S_IRUGO, env_show_keyswitch, NULL);
static ssize_t env_show_fspleds(struct device *d, struct device_attribute *attr, char *buf);
/* we can set LEDs, but until we have bus-atomicity it is racy */
static ssize_t env_store_fspleds(struct device *d, struct device_attribute *attr, const char *b, size_t c);
static DEVICE_ATTR(fspleds, \
	S_IRUGO|S_IWUSR, env_show_fspleds, env_store_fspleds);
#if ENV_ENABLE_INTR
static void env_handle_ps_statechange_tasklet(unsigned long data);
static void env_handle_keyswitch_tasklet(unsigned long data);
static void env_handle_fan_statechange_tasklet(unsigned long data);
#endif /* ENV_ENABLE_INTR */
static void env_update_fanstat(void);
static char* env_keypos_to_str(u8 kpos);
#if ENV_DEBUG_RAW
	static ssize_t env_show_raw (struct device *d, struct device_attribute *attr, char *buf);
	static DEVICE_ATTR(raw, S_IRUGO, env_show_raw, NULL);
#endif /* ENV_DEBUG_RAW */

// declare tasklets
#if ENV_ENABLE_INTR
DECLARE_TASKLET(ps_statechange_tasklet, env_handle_ps_statechange_tasklet, 0);
DECLARE_TASKLET(keyswitch_tasklet, env_handle_keyswitch_tasklet, 0);
DECLARE_TASKLET(fan_statechange_tasklet, env_handle_fan_statechange_tasklet, 0);
/*
   These are for identifying which PS has change state when
   the tasklet is called.
*/
static bool ps0_changed_state;
static bool ps1_changed_state;
static bool ps2_changed_state;
spinlock_t ps_changed_state_lock;
#endif /* ENV_ENABLE_INTR */

// we need to track how many times each LED is set, so we know when each
// is safe to extinguish
static int generr_setters = 0;
static int temperr_setters = 0;
// we need to lock them, too
spinlock_t fspled_lock;

u8 env_get_psu_state(u8 psu, u8 *state);

/* specify "envctrl_overtemp=" to override the default over-temperature
 * action.  Allowable values are poweroff, shutdown and none (dangerous!).
 * The default is shutdown to be idiomatic with existing Linux/SPARC envctrl
 * drivers.
 * They all ought to be "poweroff" for safety-- though this requires
 * tweaking the scons_poweroff sysctl on kernels that support it,
 * otherwise serial console based systems will not poweroff properly.
 */
static int __init env_parse_overtemp_option(const char *str)
{
	if (NULL == str) {
		return 1;
	}
 
	if (!strncmp(str, "poweroff", strlen("poweroff"))) {
		env_poweroff_pfn = env_do_poweroff;
		return 1;
	}
	else if (!strncmp(str, "shutdown", strlen("shutdown"))) {
		env_poweroff_pfn = env_do_shutdown;
		return 1;
	}
#if ENV_DEBUG_TEMP
	else if (!strncmp(str, "none", strlen("none"))) {
		env_poweroff_pfn = NULL;
		printk(KERN_INFO 
			"%s: WARNING: thermal shutdown/power-off disabled!\n",
			ENV_DEVNAME);
		return 1;
	}
#endif
	else {
		printk(KERN_ERR 
			"%s: unrecognized envctrl_overtemp value \"%s\" ignored\n", 
			ENV_DEVNAME, str);
	}
	return 0;
}

static char *envctrl_overtemp;
module_param(envctrl_overtemp, charp, 0);
MODULE_PARM_DESC(envctrl_overtemp,
        "over-temperature action: poweroff or shutdown (default)");

static int envctrl_mon_interval = 60;
module_param(envctrl_mon_interval, uint, 0);
MODULE_PARM_DESC(envctrl_mon_interval,
        "environmental monitoring interval (seconds)");

static int __init env_init_lm75(struct therm *therm)
{
	u16 dat;

	/* 
	 * Arbitrary LM75 initialization 
	 * 
	 * The LM75 OS# (overtemp shutdown) does not seem to be
	 * implemented, so setting shutdown and hysteresis values
	 * isn't really required...
	 */

	/* Tos is +5 degC to give OS a chance at shutdown */
	dat = (therm->cfg->shutdown_temp + 5) << 8; 
	if (envctrl_i2c_write_LM75(
		therm->cfg->addr, ENV_LIB_LM75_REG_TOS, dat)) {
			printk(KERN_ERR "%s: failed to init LM75 Tos %i C\n",
				ENV_DEVNAME, therm->cfg->shutdown_temp);
			return 1;
	}
	else {
		printk(KERN_DEBUG "%s: initialized LM75 Tos %i C\n",
			ENV_DEVNAME, therm->cfg->shutdown_temp);
	}

	/* arbitrary -5 degC from Tos value for hysteresis */
	dat = (therm->cfg->shutdown_temp) << 8; 
	if (envctrl_i2c_write_LM75(
		therm->cfg->addr, ENV_LIB_LM75_REG_THYST, dat)) {
			printk(KERN_ERR "%s: failed to init LM75 Thyst %i C\n",
				ENV_DEVNAME, therm->cfg->shutdown_temp);
			return 1;
	}
	else {
		printk(KERN_DEBUG "%s: initialized LM75 Thyst %i C\n",
			ENV_DEVNAME, therm->cfg->shutdown_temp - 5);
	}

	/* Set fault-queue depth of 2 -- perhaps not necessary
	 */

	dat = ENV_LM75_INIT;
	if (envctrl_i2c_write_LM75(
		therm->cfg->addr, ENV_LIB_LM75_REG_CFG, dat)) {
			printk(KERN_ERR "%s: failed to init LM75 cfg 0x%04x\n",
				ENV_DEVNAME, dat);
			return 1;
	}
	else {
		printk(KERN_DEBUG "%s: initialized LM75 cfg 0x%04x\n",
			ENV_DEVNAME, dat);
	}

	return 0;
}

static inline struct property * env_get_property(const char *name, int *lenp)
{
	if (lenp) {
		*lenp = -1;
	}
	return of_find_property(nex->ofdev->node, name, lenp);
}

static int __init env_init_thermisters(void)
{
	int i;
	int ntherms = 0;
	struct env_thermister_prop *t_itr;
	u8 *start, *end;

	struct therm *therm;
	struct list_head *tmp1, *tmp2;	

	struct property *pp;

	/*
	 * The lookup table length is short by two entries that 
	 * represent the coolest thermister raw values 
	 * (indecies 254, 255).  These lower values are reasonably
	 * expected, so we allocate a bigger table and shift the table
	 * values down by two.  We'll duplicate the value at the top of
	 * the table for the two new slots (though these are cooking
	 * temperatures we'd not expect to see).
	 */
	pp = env_get_property("cpu-temp-factors", &i);
	if (i == 254) {
		/* TODO: memset is for good measure-- likely not necessary */
		memset(&env_cpu_temp_table, '\0', sizeof(env_cpu_temp_table));

		/* TODO: we should use the in-memory table and index -2 */
		memcpy((u8*)&env_cpu_temp_table +2, pp->value, i);
		env_cpu_temp_table[0] = \
			env_cpu_temp_table[1] = env_cpu_temp_table[2];
	}
	else {
		if (i != -1) {
			printk(KERN_WARNING
				"%s: unexpected cpu-temp-factors table size " \
				"(is %d, expected %d)\n", ENV_DEVNAME, i, 254);
		}
		else {
			/* This is the first indication our OBP does not create
		 	* all desired properties-- inform the user...
		 	* We could parse the OBP version, but why bother?
		 	* TODO: Assume OBP >= 3.12 provides entries-- check it!
		 	*/
			printk(KERN_WARNING
				"%s: OBP properties incomplete-- "
				"consider updating to OBP >= 3.12\n",
				ENV_DEVNAME);
#if ENV_OBP_IGN_CPUS
			printk(KERN_WARNING
				"%s: only CPU0 will be monitored\n",
				ENV_DEVNAME);
#endif
		}
		printk(KERN_INFO 
			"%s: using default cpu-temp-factors table\n", 
			ENV_DEVNAME);
	}

	pp = env_get_property("cpu-fan-speeds", &i);
	if (i != ENV_FANTBL_SZ) {
		printk(KERN_ERR "%s: unexpected cpu-fan-speeds table size " \
			"(is %u expected %u)\n", ENV_DEVNAME, i, ENV_FANTBL_SZ);
		goto out_err;
	}

	// TODO: we should use the in-memory copy	
	memcpy((u8*)&env_cpu_fanspeed_table, pp->value, i);

	pp = env_get_property("ps-temp-factors", &i);
	if (i == 254) {
		/* TODO: memset is for good measure-- likely not necessary */
		memset(&env_ps_temp_table, '\0', sizeof(env_ps_temp_table));

		/* TODO: use in-memory copy */
		memcpy((u8*)&env_ps_temp_table +2, pp->value, i);
		env_ps_temp_table[0] = \
			env_ps_temp_table[1] = env_ps_temp_table[2];
	}
	else {
		if (i != -1) {
			printk(KERN_WARNING
				"%s: unexpected ps-temp-factors table size " \
				"(is %d, expected %d)\n", ENV_DEVNAME, i, 254);
		}
		printk(KERN_INFO 
			"%s: using default ps-temp-factors table\n", 
			ENV_DEVNAME);
	}

	pp = env_get_property("ps-fan-speeds", &i);
	if (i != ENV_FANTBL_SZ) {
		printk(KERN_ERR "%s: unexpected ps-fan-speeds table size " \
			"(is %u expected %u)\n", ENV_DEVNAME, i, ENV_FANTBL_SZ);
		goto out_err;
	}
	memcpy((u8*)&env_ps_fanspeed_table, pp->value, i);	

	pp = env_get_property("thermisters", &i);
	if (i == -1) {
		printk(KERN_INFO "%s: using default thermister data\n",
			ENV_DEVNAME);
		env_therms_data = env_therms_data_default;
		i = sizeof(env_therms_data_default);
		
	}
	else {
		env_therms_data = pp->value;
	}	
	start = env_therms_data;
	end   = start + i - 1;

	while (start < end &&
	      ((end - start) >= sizeof(struct env_thermister_prop) - 1)) {
		t_itr = (struct env_thermister_prop *) start;
		start = t_itr->name + strlen(t_itr->name) + 1;

		therm = kmalloc(sizeof(struct therm), GFP_KERNEL);
		if (!therm) {
			printk(KERN_ERR
				"%s: unable to allocate thermister\n",
				ENV_DEVNAME);
			goto out_err;
		}

		/* init one thermister */
		therm->cfg       = t_itr;
		therm->disabled  = 0;
		therm->last_warn = 0;

		if (!strncmp(therm->cfg->name, "CPU", strlen("CPU"))) {
			therm->type = therm_type_cpu;
		}
		else if (!strncmp(therm->cfg->name, "PS", strlen("PS"))) {
			therm->type = therm_type_noncpu;

			/* TODO: use interrupt handling to determine
			 * PSU add/removal-- for now, this is determined
			 * once (expensively) at module load time (FIXME!)
			 */
			{
				u8 psu = therm->cfg->name[2] - '0';
				u8 psustat;
				env_get_psu_state(psu, &psustat);
				if (! ENV_PS_PRES(psustat)) {
					therm->disabled = 1;
					printk(KERN_DEBUG 
						"%s: PS%i not present-- "
						"monitoring disabled\n",
						ENV_DEVNAME, psu);
				}
			}	
		}
		else if (!strncmp(therm->cfg->name, "AMB", strlen("AMB"))) {
			therm->type = therm_type_lm75;
			(void) env_init_lm75(therm);
		}
		else {
			printk(KERN_ERR "%s: unknown thermister type \"%s\""
				"-- disabled\n",
				ENV_DEVNAME, therm->cfg->name);
			therm->disabled = 1;
			therm->type     = therm_type_unk;
			continue;
		}

		list_add_tail(&(therm->list), &env_therms);
		ntherms++;
	}
	return ntherms;

#if ENV_ENABLE_INTR
	// init our PS state-change spinlock for use later
	spin_lock_init(&ps_changed_state_lock);
#endif

out_err:
	list_for_each_safe(tmp1, tmp2, &env_therms) {
		therm = list_entry(tmp1, struct therm, list);
		list_del(tmp1);
		kfree(therm);
	}
	return 0;
}

#if ENV_ENABLE_INTR
static irqreturn_t env_fsp_intr(int irq, void *dev_id)
{
	/* 
	 * Known interrupt sources:
	 * 0xDF - Keyswitch	(0x20)
	 * 0xEF - Fan		(0x10)
	 * 0xFB - PS0		(0x04)
	 * 0xFD - PS1		(0x02)
	 * 0xFE - PS2		(0x01)
	 */

	u8 srcs = envctrl_i2c_read_8574(ENV_FSP_CTRL_ADDR);
	int known = 0;
	unsigned long flags;
	u8 newstate = srcs;

#if ENV_DEBUG_INTR
	printk(KERN_DEBUG "%s: %s: interrupt 0x%02x (masked 0x%02x)\n",
		ENV_DEVNAME, __FUNCTION__, srcs, srcs ^ env_pstate);
#endif /* ENV_DEBUG_INTR */

	/*
	 * Handle events
	 */

	if ((srcs ^ env_pstate) & 0x20) {
		tasklet_schedule(&keyswitch_tasklet);
		srcs |= 0x20;
		known = 1;
	}

	if ((srcs ^ env_pstate) & 0x10) {
		tasklet_schedule(&fan_statechange_tasklet);
		srcs |= 0x10;
		known = 1;
	}

	if ((srcs ^ env_pstate) & 0x04) {
		spin_lock_irqsave(&ps_changed_state_lock, flags);
		ps0_changed_state = 1;
		spin_unlock_irqrestore(&ps_changed_state_lock, flags);
		tasklet_schedule(&ps_statechange_tasklet);
		srcs |= 0x04;
		known = 1;
	}

	if ((srcs ^ env_pstate) & 0x02) {
		spin_lock_irqsave(&ps_changed_state_lock, flags);
		ps1_changed_state = 1;
		spin_unlock_irqrestore(&ps_changed_state_lock, flags);
		tasklet_schedule(&ps_statechange_tasklet);
		srcs |= 0x02;
		known = 1;
	}

	if ((srcs ^ env_pstate) & 0x01) {
		spin_lock_irqsave(&ps_changed_state_lock, flags);
		ps2_changed_state = 1;
		spin_unlock_irqrestore(&ps_changed_state_lock, flags);
		tasklet_schedule(&ps_statechange_tasklet);
		srcs |= 0x01;
		known = 1;
	}
	
	if (!known && ((srcs ^ env_pstate) != 0))
		printk(KERN_DEBUG "%s: unhandled interrupt 0x%02x, masked to 0x%02x\n",
			ENV_DEVNAME, srcs, srcs ^ env_pstate);

	env_pstate = newstate;
	return IRQ_HANDLED;
}

static void env_handle_ps_statechange_tasklet(unsigned long data) {
	struct therm *t;
	struct list_head *tmp1, *tmp2;
	unsigned long flags;

	spin_lock_irqsave(&ps_changed_state_lock, flags);
	
	list_for_each_safe(tmp1, tmp2, &env_therms) {
		t = list_entry(tmp1, struct therm, list);
		if (t->type == therm_type_noncpu) {
			if (((!strcmp(t->cfg->name, "PS0")) && (ps0_changed_state == 1)) ||
			    ((!strcmp(t->cfg->name, "PS1")) && (ps1_changed_state == 1)) ||
			    ((!strcmp(t->cfg->name, "PS2")) && (ps2_changed_state == 1))) {
				if (t->disabled == 0) {
					t->disabled = 1;
					printk(KERN_NOTICE "%s: %s removed from system -- monitoring disabled\n",
						ENV_DEVNAME, t->cfg->name);
				} else {
					t->disabled = 0;
					printk(KERN_NOTICE "%s: %s inserted into system -- monitoring enabled\n",
						ENV_DEVNAME, t->cfg->name);
				}
			}
		}
	}
	
	ps0_changed_state = 0;
	ps1_changed_state = 0;
	ps2_changed_state = 0;
	
	spin_unlock_irqrestore(&ps_changed_state_lock, flags);
}
#endif /* ENV_ENABLE_INTR */

static u8 env_get_keyswitch(void)
{
	return envctrl_i2c_read_8574(ENV_FSP_LED_ADDR) & ENV_FSP_KEYMSK;
}

#if ENV_ENABLE_INTR
static void env_handle_keyswitch_tasklet(unsigned long data)
{
	/* TODO: handle each keyswitch setting */
	u8 kpos;
	kpos = env_get_keyswitch();

	printk(KERN_DEBUG "%s: keyswitch position is %s\n",
		ENV_DEVNAME, env_keypos_to_str(kpos));

	switch (kpos) {
		case ENV_FSP_KEY_OFF:
			/* Keyswitch is set to OFF??? */
			printk(KERN_ERR "%s: keyswitch is set to off!\n",
				ENV_DEVNAME);
			break;
		case ENV_FSP_KEY_DIAG:
			/* Keyswitch is set to DIAG */
			break;
		case ENV_FSP_KEY_ON:
			/* Keyswitch is set to ON */
			break;
		case ENV_FSP_KEY_LOCK:
			/* Keyswitch is set to KEY_LOCK */
			break;
		default:
			/* Keyswitch is set to unknown position! */
			printk(KERN_ERR "%s: keyswitch is set to unknown position!\n",
				ENV_DEVNAME);
			break;
	}
}

static void env_handle_fan_statechange_tasklet(unsigned long data)
{
	env_update_fanstat();
}

#endif /* ENV_ENABLE_INTR */

static void env_fsp_activity(unsigned long unused)
{
	u8 rd;
	unsigned long flags;

	/* TODO: this must be made i2c-atomic-- not just spinlocked */
	/* FIXME: do we actually need to irqsave here? */
	spin_lock_irqsave(&nex->lock, flags);

	__envctrl_i2c_read_8574(ENV_FSP_LED_ADDR);
	/* always write keyswitch mask-- key bits can be unset, but not set! */
	rd = (__envctrl_i2c_read_8574(ENV_FSP_LED_ADDR) ^ ENV_LED_ACTIVE) 
		| ENV_FSP_KEYMSK;
	__envctrl_i2c_write_8574(ENV_FSP_LED_ADDR, rd);

	spin_unlock_irqrestore(&nex->lock, flags);

	mod_timer(&env_fsp_act_timer, jiffies + (HZ/2));
}

/* TODO: change to include return value */
static void env_set_fanspeed(struct fan *f, u8 fanspeed, u8 forced)
{
	if (!forced) {
		if (f->forced == 1) {
			return;
		}
	}
	if (fanspeed > ENV_FAN_MAX) {
		fanspeed = ENV_FAN_MAX;
	}
	
	if (fanspeed != f->last_speed) {
		f->last_speed = fanspeed;
		envctrl_i2c_write_8444(ENV_FAN_ADDR, f->fanport, fanspeed);
	}
}

static int __init env_init_fsp(void) 
{
	int i;
	struct property *pp;

	if (nex->ofdev->num_irqs != 2) {
		printk(KERN_ERR "%s: unexpected number of interrupts "
			"(is %i expected %i)\n", 
			ENV_DEVNAME, nex->ofdev->num_irqs, 2);
		return 0;
	}
#if ENV_ENABLE_INTR
	env_fsp_irq = nex->ofdev->irqs[1];
	if (request_irq(env_fsp_irq, &env_fsp_intr,
			IRQF_SHARED, "SUNW,envctrl FSP",
			(void *)&env_fsp_act_timer)) {
		printk(KERN_ERR "%s: unable to register FSP irq\n", 
			ENV_DEVNAME);
		env_fsp_irq = 0;
		return 0;
	}
#endif /* ENV_ENABLE_INTR */
	/* 
	 * FSP GEN_ERR LED is plumbed OR'ed with auxio LED bit.
	 * Extinguish it, otherwise system appears in error
	 */
	auxio_set_led(AUXIO_LED_OFF);

	/* 
	 * Reset FSP LED/Key bits-- 
	 * this extinguishes all LEDs and resets keyswitch reading
	 */
	envctrl_i2c_write_8574(ENV_FSP_LED_ADDR, 
		ENV_FSP_LEDMSK | ENV_FSP_KEYMSK);

	init_timer(&env_fsp_act_timer);
	env_fsp_act_timer_init = 1;
	env_fsp_act_timer.expires  = jiffies + (HZ/2);
	env_fsp_act_timer.function = env_fsp_activity;
	add_timer(&env_fsp_act_timer);

	/* This is a funny place to diddle the fans, but we don't want
	 * an unprogrammed DAC to allow fans to drop to 0 speed once 
	 * they are "unlocked" via the DFLOP writes below.
	 * TODO: put me somewhere else!
	 */
	env_set_fanspeed(&env_cpu_fans, 0x30, 0); /* 0x30 is arbitrary */
	env_set_fanspeed(&env_psu_fans,  0x30, 0); /* 0x30 is arbitrary */
	/* AFB fan not present on later models-- be safe and set to max */
	env_set_fanspeed(&env_afb_fan, ENV_FAN_MAX, 0);

	// set status of fan assemblies
	env_update_fanstat();

	/* Enable FSP interrupts */
	envctrl_i2c_write_8574(ENV_FSP_CTRL_ADDR, ENV_FSP_DFLOP_INIT0);
	udelay(10); /* debug */
	envctrl_i2c_write_8574(ENV_FSP_CTRL_ADDR, ENV_FSP_DFLOP_INIT1);
	udelay(10); /* debug */
	envctrl_i2c_write_8574(ENV_FSP_CTRL_ADDR, ENV_FSP_DEVINTR_INIT0);
	udelay(10); /* debug */
	envctrl_i2c_write_8574(ENV_FSP_CTRL_ADDR, ENV_FSP_DEVINTR_INIT1);
	udelay(10); /* debug */

	/* Initialize disk backplane LED related bits */
	/* TODO: move me somewhere else */
	pp = env_get_property("disk-leds-present", &i);
	if (i != sizeof(env_disk_bp)) {
		printk(KERN_WARNING
			"%s: unexpected disk-leds-present size " \
			"(is %d, expected %lu)\n", 
			ENV_DEVNAME, i, sizeof(env_disk_bp));
		return 0;
	}
	memcpy((u8*)&env_disk_bp, pp->value, i);
	for (i = 0; i < ENV_DISK_BP_MAX; ++i) {
		printk(KERN_DEBUG "%s: disk backplane %i: %s\n", 
			__FUNCTION__, i,
			(!env_disk_bp[i].mask) ? "empty" :
			  ((env_disk_bp[i].mask == ENV_DISK_BP_8SLOT) ?
			    "8-slot" : (
			      (env_disk_bp[i].mask == ENV_DISK_BP_4SLOT) ? 
			        "4-slot" : "unknown")));
	}

	return 1;
}

#define ENV_FAN_WDT_ADDR (0xA0)
/* TODO: This function should set the timer to 2*envctrl_mon_interval
 * using seconds or minutes as appropriate-- it is ack'ed by kenvctrld
 */
static int __init env_init_fan_wdt(void) 
{
	/* 
	 * Initialize PCF8583 fan watchdog timer--
	 * This device likely kicks fans into max speed when INT#
	 * is asserted.  We use the countdown timer functionality of
	 * the PCF8583 as a watchdog-- if we don't "kick" it, the
	 * fans blast.  This is protection should the kernel get wedged.
	 */

	/* disable counting */
	if (envctrl_i2c_write_pcf8583(
		ENV_FAN_WDT_ADDR, ENV_LIB_8583_REG_CSR, (1<<7))) {

		printk(KERN_ERR "%s: PCF8583: failed to disable counting\n",
			ENV_DEVNAME);
		return 0;
	}

	/* reset timer register (reg 0x07) */
	if (envctrl_i2c_write_pcf8583(
		ENV_FAN_WDT_ADDR, ENV_LIB_8583_REG_TIMER, 0)) {

		printk(KERN_ERR "%s: PCF8583: failed to reset timer\n",
			ENV_DEVNAME);
		return 0;
	}

	/* set 60 (seconds) into alarm reg  (reg 0x0f) */
	/* TODO: this is not yet working-- we trip on overflow currently */
	if (envctrl_i2c_write_pcf8583(
		ENV_FAN_WDT_ADDR, ENV_LIB_8583_REG_ALRMTMR, 60)) {

		printk(KERN_ERR "%s: PCF8583: failed to set alarm timer\n",
			ENV_DEVNAME);
		return 0;
	}

	/* enable alarm interrupts, counter counts in seconds */
	if (envctrl_i2c_write_pcf8583(
		ENV_FAN_WDT_ADDR, ENV_LIB_8583_REG_ALRMCTL, 0xCA)) {

		printk(KERN_ERR "%s: PCF8583: failed initialize alarm ctl\n",
			ENV_DEVNAME);
		return 0;
	}

	/* (1<<2) = Alarm Enable Bit - stops default 1Hz INT#, enables cntr */
	if (envctrl_i2c_write_pcf8583(
		ENV_FAN_WDT_ADDR, ENV_LIB_8583_REG_CSR, (1<<2))) {

		printk(KERN_ERR "%s: PCF8583: failed to enable alarm\n",
			ENV_DEVNAME);
		return 0;
	}

	return 1;
}

/* resets the watchdog timer 
 * TODO: check if wdog is tripped and take corrective action (or report)
 */
static int env_ack_fan_wdt(void) 
{
	/* 
	 * Reset PCF8583 fan watchdog timer
	 * TODO: stop counter first for safety???
	 */
	if (envctrl_i2c_write_pcf8583(
		ENV_FAN_WDT_ADDR, ENV_LIB_8583_REG_TIMER, 0)) {

		printk(KERN_ERR "%s: PCF8583: failed to ack fan watchdog\n",
			ENV_DEVNAME);
		return 0;
	}
	return 1;
}

static u8 env_read_thermister(struct therm *t, u8 *raw)
{
	u8 val;
	int ret;

	switch (t->cfg->addr & 0xF0) {
		case 0x90:
			if (t->type == therm_type_lm75) {
				/* LM75 shares PCF8591 address space */
				ret = envctrl_i2c_read_LM75_temp(
					t->cfg->addr, &val);
			}
			else {
				ret = envctrl_i2c_read_8591(
					t->cfg->addr, t->cfg->port, &val);
			}
			break;
		default:
			/* BUG() */
			return 0;
	}

	if (ret) {
		printk(KERN_ERR "%s: unable to read thermister %s%s\n",
			ENV_DEVNAME, t->cfg->name,
			(ret == -EINTR)?" (lost arbitration-- transient)":"");
		return 0;
	}
	if (raw) {
		*raw = val;
	}

	/* I'm not sure of the intention, as num and den don't seem
	 * to be used (both set to 1) in this implementation. I could 
	 * guess they take precedence over table lookups if non-zero 
	 * and non-one (or they apply after an appropriate table lookup 
	 * under all circumstances???).
	 */
	if (t->cfg->num > 1 && t->cfg->den > 1) {
		val = (val * t->cfg->num) / t->cfg->den;
	}
	else {
		switch (t->type) 
		{
			case therm_type_cpu:
				val = env_cpu_temp_table[val];
				break;
			case therm_type_noncpu:
				val = env_ps_temp_table[val];
				break;
			case therm_type_lm75:
				/* untranslated */
				break;
			case therm_type_unk:
				/* fall-thru */
			default:
				/* no conversion */
				break;
		}
	}
	return val;
}

/* TODO: check/provide meaningful return value */
static u8 env_get_fanspeed(struct fan *f)
{
	u8 rv;
	(void) envctrl_i2c_read_8591(ENV_FAN_SPD_ADDR ,f->fanport ,&rv); 
	rv >>= 2;
	/* TODO: adjust value by ~ +7% as per E250 implementation ? */
	return rv;
}

/* TODO: diagnostic return value */
static void env_set_fspleds(u8 leds_on, u8 leds_off)
{
	u8 rd;
	unsigned long flags;

	/* don't stomp on keyswitch bits */
	if ((leds_on | leds_off) & ENV_FSP_KEYMSK) {
		return;
	}

	/* TODO: this must be made i2c-atomic-- not just spinlocked */
	spin_lock_irqsave(&nex->lock, flags);

	__envctrl_i2c_read_8574(ENV_FSP_LED_ADDR);
	/* always write keyswitch mask-- key bits can be unset, but not set! */
	rd = (__envctrl_i2c_read_8574(ENV_FSP_LED_ADDR) | 
		ENV_FSP_KEYMSK | leds_off) & ~leds_on;
	__envctrl_i2c_write_8574(ENV_FSP_LED_ADDR, rd);

	spin_unlock_irqrestore(&nex->lock, flags);
}

static void env_update_fanstat(void)
{
	u8 raw_stat = envctrl_i2c_read_8574(ENV_FAN_STAT_ADDR);
	u8 new_raw_state;
	bool last_fanstate;
	unsigned long flags;
	u8 ledtemp = 0;

	// TODO: we should probably store fans in a list...
	// FIXME: there's lots of duplicated code here...
	
	new_raw_state = ENV_FAN_ERR_CPU(raw_stat);
	
	spin_lock_irqsave(&fan_states_lock, flags);
	spin_lock(&fspled_lock);
	
	last_fanstate = env_cpu_fans_state.fan0state;
	if (new_raw_state & ENV_FAN_STAT_CPU0) {
		env_cpu_fans_state.fan0state = 1;
		if (last_fanstate != 1) {
			printk(KERN_ERR "%s: the CPU fan assembly reports failure of fan0!\n",
				ENV_DEVNAME);
			// set GENERR and TEMPERR LEDs
			generr_setters++;
			temperr_setters++;
			env_set_fspleds((ENV_LED_GENERR ^ ENV_LED_TEMPERR), 0);
		}
	} else {
		env_cpu_fans_state.fan0state = 0;
		if (last_fanstate != 0) {
			printk(KERN_NOTICE "%s: the CPU fan assembly failure of fan0 has cleared.\n",
				ENV_DEVNAME);
			// clear LEDs (if possible)
			generr_setters--;
			temperr_setters--;
			if ((generr_setters == 0) && (temperr_setters == 0)) {
				ledtemp = (ENV_LED_GENERR ^ ENV_LED_TEMPERR);
			} else if (temperr_setters == 0) {
				ledtemp = ENV_LED_TEMPERR;
			}
			env_set_fspleds(0, ledtemp);
		}
	}
	
	last_fanstate = env_cpu_fans_state.fan1state;
	if (new_raw_state & ENV_FAN_STAT_CPU1) {
		env_cpu_fans_state.fan1state = 1;
		if (last_fanstate != 1) {
			printk(KERN_ERR "%s: the CPU fan assembly reports failure of fan1!\n",
				ENV_DEVNAME);
			// set GENERR and TEMPERR LEDs
			generr_setters++;
			temperr_setters++;
			env_set_fspleds((ENV_LED_GENERR ^ ENV_LED_TEMPERR), 0);
		}
	} else {
		env_cpu_fans_state.fan1state = 0;
		if (last_fanstate != 0) {
			printk(KERN_NOTICE "%s: the CPU fan assembly failure of fan1 has cleared.\n",
				ENV_DEVNAME);
			// clear LEDs (if possible)
			generr_setters--;
			temperr_setters--;
			if ((generr_setters == 0) && (temperr_setters == 0)) {
				ledtemp = (ENV_LED_GENERR ^ ENV_LED_TEMPERR);
			} else if (temperr_setters == 0) {
				ledtemp = ENV_LED_TEMPERR;
			}
			env_set_fspleds(0, ledtemp);
		}
	}

	last_fanstate = env_cpu_fans_state.fan2state;
	if (new_raw_state & ENV_FAN_STAT_CPU2) {
		env_cpu_fans_state.fan2state = 1;
		if (last_fanstate != 1) {
			printk(KERN_ERR "%s: the CPU fan assembly reports failure of fan2!\n",
				ENV_DEVNAME);
			// set GENERR and TEMPERR LEDs
			generr_setters++;
			temperr_setters++;
			env_set_fspleds((ENV_LED_GENERR ^ ENV_LED_TEMPERR), 0);
		}
	} else {
		env_cpu_fans_state.fan2state = 0;
		if (last_fanstate != 0) {
			printk(KERN_NOTICE "%s: the CPU fan assembly failure of fan2 has cleared.\n",
				ENV_DEVNAME);
			// clear LEDs (if possible)
			generr_setters--;
			temperr_setters--;
			if ((generr_setters == 0) && (temperr_setters == 0)) {
				ledtemp = (ENV_LED_GENERR ^ ENV_LED_TEMPERR);
			} else if (temperr_setters == 0) {
				ledtemp = ENV_LED_TEMPERR;
			}
			env_set_fspleds(0, ledtemp);
		}
	}

	new_raw_state = ENV_FAN_ERR_PS(raw_stat);

	last_fanstate = env_psu_fans_state.fan0state;
	if (new_raw_state & ENV_FAN_STAT_PS0) {
		env_psu_fans_state.fan0state = 1;
		if (last_fanstate != 1) {
			printk(KERN_ERR "%s: the PSU fan assembly reports failure of fan0!\n",
				ENV_DEVNAME);
			// set GENERR and TEMPERR LEDs
			generr_setters++;
			temperr_setters++;
			env_set_fspleds((ENV_LED_GENERR ^ ENV_LED_TEMPERR), 0);
		}
	} else {
		env_psu_fans_state.fan0state = 0;
		if (last_fanstate != 0) {
			printk(KERN_NOTICE "%s: the PSU fan assembly failure of fan0 has cleared.\n",
				ENV_DEVNAME);
			// clear LEDs (if possible)
			generr_setters--;
			temperr_setters--;
			if ((generr_setters == 0) && (temperr_setters == 0)) {
				ledtemp = (ENV_LED_GENERR ^ ENV_LED_TEMPERR);
			} else if (temperr_setters == 0) {
				ledtemp = ENV_LED_TEMPERR;
			}
			env_set_fspleds(0, ledtemp);
		}
	}
	
	last_fanstate = env_psu_fans_state.fan1state;
	if (new_raw_state & ENV_FAN_STAT_PS1) {
		env_psu_fans_state.fan1state = 1;
		if (last_fanstate != 1) {
			printk(KERN_ERR "%s: the PSU fan assembly reports failure of fan1!\n",
				ENV_DEVNAME);
			// set GENERR and TEMPERR LEDs
			generr_setters++;
			temperr_setters++;
			env_set_fspleds((ENV_LED_GENERR ^ ENV_LED_TEMPERR), 0);
		}
	} else {
		env_psu_fans_state.fan1state = 0;
		if (last_fanstate != 0) {
			printk(KERN_NOTICE "%s: the PSU fan assembly failure of fan1 has cleared.\n",
				ENV_DEVNAME);
			// clear LEDs (if possible)
			generr_setters--;
			temperr_setters--;
			if ((generr_setters == 0) && (temperr_setters == 0)) {
				ledtemp = (ENV_LED_GENERR ^ ENV_LED_TEMPERR);
			} else if (temperr_setters == 0) {
				ledtemp = ENV_LED_TEMPERR;
			}
			env_set_fspleds(0, ledtemp);
		}
	}

	last_fanstate = env_psu_fans_state.fan2state;
	if (new_raw_state & ENV_FAN_STAT_PS2) {
		env_psu_fans_state.fan2state = 1;
		if (last_fanstate != 1) {
			printk(KERN_ERR "%s: the PSU fan assembly reports failure of fan2!\n",
				ENV_DEVNAME);
			// set GENERR and TEMPERR LEDs
			generr_setters++;
			temperr_setters++;
			env_set_fspleds((ENV_LED_GENERR ^ ENV_LED_TEMPERR), 0);
		}
	} else {
		env_psu_fans_state.fan2state = 0;
		if (last_fanstate != 0) {
			printk(KERN_NOTICE "%s: the PSU fan assembly failure of fan2 has cleared.\n",
				ENV_DEVNAME);
			// clear LEDs (if possible)
			generr_setters--;
			temperr_setters--;
			if ((generr_setters == 0) && (temperr_setters == 0)) {
				ledtemp = (ENV_LED_GENERR ^ ENV_LED_TEMPERR);
			} else if (temperr_setters == 0) {
				ledtemp = ENV_LED_TEMPERR;
			}
			env_set_fspleds(0, ledtemp);
		}
	}
	
	spin_unlock(&fspled_lock);
	spin_unlock_irqrestore(&fan_states_lock, flags);

	// TODO: check AFB fan?

	printk(KERN_DEBUG "%s: fan stat 0x%02x\n", 
		__FUNCTION__, raw_stat);
}

static ssize_t env_show_fanspeed(struct device *dev, struct device_attribute *attr, char *buf)
{
	ssize_t len;
	len =  sprintf(buf, "CPU FANS:\n  Speed: %u\n  Speed forced: %s\n  Health: fan0 = %s, fan1 = %s, fan2 = %s\n", 
		env_get_fanspeed(&env_cpu_fans), 
		env_cpu_fans.forced ? "YES" : "NO",
		env_cpu_fans_state.fan0state ? "FAILED" : "OK",
		env_cpu_fans_state.fan1state ? "FAILED" : "OK",
		env_cpu_fans_state.fan2state ? "FAILED" : "OK");
	len +=  sprintf(buf + len, "PS FANS:\n  Speed: %u\n  Speed forced: %s\n  Health: fan0 = %s, fan1 = %s, fan2 = %s\n", 
		env_get_fanspeed(&env_psu_fans), 
		env_psu_fans.forced ? "YES" : "NO",
		env_psu_fans_state.fan0state ? "FAILED" : "OK",
		env_psu_fans_state.fan1state ? "FAILED" : "OK",
		env_psu_fans_state.fan2state ? "FAILED" : "OK");
	len += sprintf(buf + len, "AFB FAN:\n  Speed: %u (???)\n  Speed forced: %s\n  Health: IGNORED\n",
		env_get_fanspeed(&env_afb_fan),
		env_afb_fan.forced ? "YES" : "NO");
	return len;
}

#if ENV_DEBUG_FAN
static ssize_t env_store_fanspeed(struct device *d, struct device_attribute *attr, const char *buf, size_t cnt)
{
#if 0 /* DEBUG */
	char *val = strchr(buf, '=');
	u8 spd;

	if (!val) {
		return cnt;
	}
	val++;
	
	spd = (u8) simple_strtoul(val, NULL, 0);

	if (!strnicmp(buf, "ps=", strlen("ps="))) {
		if (!spd) {
			env_psu_fans->forced = 0;
		} else {
			env_psu_fans->forced = 1;
			env_set_fanspeed(&env_psu_fans, spd, 1);
		}
	} else if (!strnicmp(buf, "cpu=", strlen("cpu="))) {
		if (!spd) {
			env_cpu_fans->forced = 0;
		} else {
			env_cpu_fans->forced = 1;
			env_set_fanspeed(&env_cpu_fans, spd, 1);
		}
	} else {
		printk("%s: unknown FAN specified (expected CPU=n or PS=n)\n",
			__FUNCTION__);
	}
#else
	/* input expected in format: port=value, port must be one char */
	u8 port = buf[0] - '0';
	u8 spd  = simple_strtoul(buf + 2, NULL, 0);
	int ret;

	// set .forced value depending on fan port being written to
	if (port == 0)
		env_cpu_fans.forced = 1;
	else if (port == 1)
		env_psu_fans.forced = 1;

	printk(KERN_DEBUG "%s: %s writing fan port %u, speed %u\n", 
		ENV_DEVNAME, __FUNCTION__, port, spd); 
	ret = envctrl_i2c_write_8444(ENV_FAN_ADDR, port, spd);
	if (ret) {
		printk(KERN_DEBUG "%s: %s failed to write fan port (%i)\n",
			ENV_DEVNAME, __FUNCTION__, ret);
	}
#endif
	return cnt;
}
#endif /* ENV_DEBUG_FAN */

#if ENV_DEBUG_TEMP
static ssize_t env_store_temp(struct device *d, struct device_attribute *attr, const char *buf, size_t count)
{
	u8 val = (u8) simple_strtoul(buf, NULL, 0);
	env_forced_temp = val;
	return count;
}

static ssize_t env_show_temp (struct device *d, struct device_attribute *attr, char *buf)
{
	return sprintf(buf, "%i\n", env_forced_temp);
}

static ssize_t env_show_lm75(struct device *d, struct device_attribute *attr, char *buf)
{
	size_t len;

	u8  lm_degC;
	u16 lm_temp;
	u16 lm_tos;
	u16 lm_thyst;
	u16 lm_cfg;

	if(envctrl_i2c_read_LM75_temp(ENV_LM75_ADDR, &lm_degC)) {
		printk(KERN_ERR "%s: failed to read lm75 temp\n", ENV_DEVNAME);
	}
	if(envctrl_i2c_read_LM75(
		ENV_LM75_ADDR, ENV_LIB_LM75_REG_TEMP, &lm_temp)) {
			printk(KERN_ERR "%s: failed to read lm75 temp reg\n",
				ENV_DEVNAME);
	}
	if(envctrl_i2c_read_LM75(
		ENV_LM75_ADDR, ENV_LIB_LM75_REG_TOS, &lm_tos)) {
			printk(KERN_ERR "%s: failed to read lm75 Tos reg\n",
				ENV_DEVNAME);
	}
	if(envctrl_i2c_read_LM75(
		ENV_LM75_ADDR, ENV_LIB_LM75_REG_THYST, &lm_thyst)) {
			printk(KERN_ERR "%s: failed to read lm75 Thyst reg\n",
				ENV_DEVNAME);
	}
	if(envctrl_i2c_read_LM75(
		ENV_LM75_ADDR, ENV_LIB_LM75_REG_CFG, &lm_cfg)) {
			printk(KERN_ERR "%s: failed to read lm75 cfg reg\n",
				ENV_DEVNAME);
	}

	len  = sprintf(buf,       "Deg:   [%i C]\n",   lm_degC);
	len += sprintf(buf + len, "TEMP:  [0x%04x]\n", lm_temp);
	len += sprintf(buf + len, "TOS:   [0x%04x]\n", lm_tos);
	len += sprintf(buf + len, "THYST: [0x%04x]\n", lm_thyst);
	len += sprintf(buf + len, "CFG:   [0x%04x]\n", lm_cfg);

	return len;	
}

static ssize_t env_store_lm75 (struct device *d, struct device_attribute *attr, const char *buf, size_t cnt)
{
	u16 temp = (u16) simple_strtoul(buf, NULL, 0);
	(void) envctrl_i2c_write_LM75(
		ENV_LM75_ADDR, ENV_LIB_LM75_REG_TOS, temp << 8);
	temp -= 5;
	(void) envctrl_i2c_write_LM75(
		ENV_LM75_ADDR, ENV_LIB_LM75_REG_THYST, temp << 8);
	return cnt;
}

static ssize_t env_store_8583 (struct device *d, struct device_attribute *attr, const char *buf, size_t cnt)
{
	u8 reg = buf[0] - '0';
	u8 val  = simple_strtoul(buf + 2, NULL, 0);
	int ret;
if (buf[0] == 'f') { reg = 0x0F; }
if (buf[0] == 'r') {
	printk(KERN_INFO "%s: resetting DFLOP/DEVINTR...\n", __FUNCTION__);
	envctrl_i2c_write_8574(ENV_FSP_CTRL_ADDR, ENV_FSP_DFLOP_INIT0);
	udelay(10); /* debug */
	envctrl_i2c_write_8574(ENV_FSP_CTRL_ADDR, ENV_FSP_DFLOP_INIT1);
	udelay(10); /* debug */
	envctrl_i2c_write_8574(ENV_FSP_CTRL_ADDR, ENV_FSP_DEVINTR_INIT0);
	udelay(10); /* debug */
	envctrl_i2c_write_8574(ENV_FSP_CTRL_ADDR, ENV_FSP_DEVINTR_INIT1);
	udelay(10); /* debug */
	return cnt;
}
if (buf[0] == 'x') {
	/* HACK-- write control device at 0x70 */
	(void) envctrl_i2c_write_8574(ENV_FSP_CTRL_ADDR, val);
	return cnt;
}
	printk(KERN_DEBUG "%s: %s writing PCF8583 reg 0x%02x, val 0x%02x\n", 
		ENV_DEVNAME, __FUNCTION__, reg, val); 
	ret = envctrl_i2c_write_pcf8583(ENV_FAN_WDT_ADDR, reg, val);
	if (ret) {
		printk(KERN_DEBUG "%s: %s failed to write PCF8583 "
			"reg 0x%02x, val 0x%02x (%i)\n",
			ENV_DEVNAME, __FUNCTION__, reg, val, ret);
	}
	return cnt;
}

static ssize_t env_show_8583(struct device *d, struct device_attribute *attr, char *buf)
{
	size_t len;

	u8 dat;
	int i;

	len = sprintf(buf, "PCF8583 registers:\n");

	for (i = 0; i <= 0x0f; i++) {
		if(envctrl_i2c_read_pcf8583(ENV_FAN_WDT_ADDR, i, &dat)) {
			printk(KERN_ERR "%s: failed to read PCF8583 reg %d\n", 
				ENV_DEVNAME, i);
			len += sprintf(buf + len, "0x%02x[%i]: <error>\n", 
				ENV_FAN_WDT_ADDR, i);
		} 
		else {
			len += sprintf(buf + len, "0x%02x[%i]: 0x%02x\n", 
				ENV_FAN_WDT_ADDR, i, dat);
		}
	}
	return len;	
}

#endif

static u8 env_suggest_fanspeed(u8 *table, u8 temp)
{
	int tblmax = ENV_FANTBL_SZ - 1; /* TODO: make ENV_FANTBL_SZ dynamic */

	/* For CPU fanspeeds, Solaris seems to index fanspeeds by [temp-20] 
	 * For PSU fanspeeds, Solaris seems to index fanspeeds by [temp-30]
	 * (as determined by Solaris prtdiag output)
	 */
	u8 offs = 0; 

	if (table == env_cpu_fanspeed_table) {
		offs = 20;
	}
	else if (table == env_ps_fanspeed_table) {
		offs = 30;
	}

	if (temp >= offs) {
		temp -= offs;
	}

	if (temp > tblmax) {
		temp = tblmax;
	}

	return *(table + temp);
}

static ssize_t env_show_thermisters(struct device *dev, struct device_attribute *attr, char *buf)
{
	ssize_t ret = 0;
	struct therm *t;
	struct list_head *tmp1, *tmp2;	

	list_for_each_safe(tmp1, tmp2, &env_therms) {
		t = list_entry(tmp1, struct therm, list);
		if (t->disabled) {
			ret += sprintf(buf + ret, "%s:\tNot present\n", t->cfg->name);
		} else {
#if ENV_DEBUG_THERMS
			{
			u8 raw;
			u8 temp = env_read_thermister(t, &raw);
			ret += sprintf(buf + ret, 
				"%s:\t%u C [%i] (0x%02x[%u] min %i, warn %i, shut %i, "
				"num %i, den %i, type %i) (sug. fan %i)\n",
				t->cfg->name, temp, raw,
				t->cfg->addr, t->cfg->port, t->cfg->min_temp, 
				t->cfg->warn_temp, t->cfg->shutdown_temp, 
				t->cfg->num, t->cfg->den, t->type,
				(t->type == therm_type_cpu ? 
					env_suggest_fanspeed(
						env_cpu_fanspeed_table,temp) :
					t->type == therm_type_noncpu ?
						env_suggest_fanspeed(
							env_ps_fanspeed_table,temp) :
							0));
			}
#else
			ret += sprintf(buf + ret, 
				"%s:\t[%u C] (min %i, warn %i, shut %i)\n",
				t->cfg->name, env_read_thermister(t, NULL),
				t->cfg->min_temp, t->cfg->warn_temp, 
				t->cfg->shutdown_temp);
#endif
		}
	}
	return ret;
}

u8 env_get_psu_state(u8 psu, u8 *state)
{
	if (psu > ENV_PS_MAX) {
		return 0;
	}
	*state = envctrl_i2c_read_8574(ENV_PS0_ADDR - (2*psu));
	return 1;
}

static ssize_t env_show_power(struct device *dev, struct device_attribute *attr, char *buf)
{
	ssize_t len = 0;
	int i;
	u8 psu_state;

	for (i = 0; i <= ENV_PS_MAX; ++i) {
		len += sprintf(buf + len, "PS%i: ", i);
		if (! env_get_psu_state(i, &psu_state)) {
			len += sprintf(buf + len, "[ERR]\n");
		}
		else {
			len += sprintf(buf +len, "[%s]\n",
				(ENV_PS_PRES(psu_state)) ? 
					"PRESENT" : "ABSENT");
		}
	}
	return len;
}

static u8 env_get_diskpres(void)
{
	return 0; /*envctrl_i2c_read_8574(ENV_DISKPRES_ADDR);*/
}

static u8 env_get_diskstat(void)
{
	return 0; /*envctrl_i2c_read_8574(ENV_DISKSTAT_ADDR);*/
}

static ssize_t env_show_disks(struct device *dev, struct device_attribute *attr, char *buf)
{
	ssize_t len;

	(void) env_get_diskpres();
	(void) env_get_diskstat(); /* XXX testing-- remove warnings */

	len = sprintf(buf, "<NOT YET IMPLEMENTED>\n");	

	if (env_disk_bp[0].mask) {
		len += sprintf(buf + len, "BP0: 0x%02x\n", 
			envctrl_i2c_read_8574(ENV_DISK_BP_ADDR_0));
	}
	if (env_disk_bp[1].mask) {
		len += sprintf(buf + len, "BP1: 0x%02x\n", 
			envctrl_i2c_read_8574(ENV_DISK_BP_ADDR_1));
	}
	if (env_disk_bp[2].mask) {
		len += sprintf(buf + len, "BP2: 0x%02x\n", 
			envctrl_i2c_read_8574(ENV_DISK_BP_ADDR_2));
	}
	return len;
}

static ssize_t env_store_disks(struct device *dev, struct device_attribute *attr, const char *buf, size_t count)
{
	u8   addr;
	u8   val;
	char *pval = strchr(buf, '=');
	if (!pval) {
		printk(KERN_DEBUG "%s: invalid input-- ignored\n",
			__FUNCTION__);
		return count;
	}
	pval++;

	val = (u8) simple_strtoul(pval, NULL, 0);
	addr = (u8) simple_strtoul(buf, NULL, 0);

	switch (addr) {
		case 0:
			if (!env_disk_bp[0].mask) {
				printk(KERN_DEBUG "%s: invalid bp specified\n",
					__FUNCTION__);
				return count;
			}
			break;
		case 1:
			if (!env_disk_bp[1].mask) {
				printk(KERN_DEBUG "%s: invalid bp specified\n",
					__FUNCTION__);
				return count;
			}
			break;
		case 2:
			if (!env_disk_bp[2].mask) {
				printk(KERN_DEBUG "%s: invalid bp specified\n",
					__FUNCTION__);
				return count;
			}
			break;
		default:
			printk(KERN_DEBUG "%s: invalid address specified\n",
				__FUNCTION__);
			return count;
	}
	addr = 0x40 + (2*addr);
	printk(KERN_DEBUG "%s: writing 0x%02x to 0x%02x\n", 
		__FUNCTION__, val, addr);
	envctrl_i2c_write_8574(addr, val);

	return count;
}

static char* env_keypos_to_str(u8 kpos)
{
	switch (kpos) {
		case ENV_FSP_KEY_OFF:
			return("OFF");
		case ENV_FSP_KEY_DIAG:
			return("DIAG");
		case ENV_FSP_KEY_ON:
			return("ON");
		case ENV_FSP_KEY_LOCK:
			return("LOCKED");
	}
	
	return ("UNKNOWN");
}

static ssize_t env_show_keyswitch(struct device *dev, struct device_attribute *attr, char *buf)
{
	ssize_t ret = 0;

	ret += sprintf(buf + ret, "KEYSWITCH: [%s]\n",
		env_keypos_to_str(env_get_keyswitch()));

	return ret;
}

static u8 env_get_fspleds(void)
{
	return envctrl_i2c_read_8574(ENV_FSP_LED_ADDR) & ENV_FSP_LEDMSK;
}

static ssize_t env_show_fspleds(struct device *dev, struct device_attribute *attr, char *buf)
{
	ssize_t ret = 0;
	u8 stat = env_get_fspleds();

	ret += sprintf(buf + ret, "POWER:      [%s]\n", "ON"); 
	ret += sprintf(buf + ret, "GEN ERROR:  [%s]\n", 
		stat & ENV_LED_GENERR ? "OFF" : "ON");
	ret += sprintf(buf + ret, "ACTIVITY:   [%s]\n", 
		stat & ENV_LED_ACTIVE ? "OFF" : "ON");
	ret += sprintf(buf + ret, "DISK ERROR: [%s]\n", 
		stat & ENV_LED_DISKERR ? "OFF" : "ON");
	ret += sprintf(buf + ret, "TEMP ERROR: [%s]\n", 
		stat & ENV_LED_TEMPERR ? "OFF" : "ON");
	ret += sprintf(buf + ret, "PS ERROR:   [%s]\n", 
		stat & ENV_LED_PSERR ? "OFF" : "ON");

	return ret;	
}

static ssize_t env_store_fspleds(struct device *dev, struct device_attribute *attr, const char *buf, size_t cnt)
{
	char on = -1;
	char led= -1;

	char *val = strchr(buf, '=');
	if (!val) {
		return cnt;
	}
	val++;
	
	if (!strnicmp(val, "on", strlen("on"))) {
		on = 1;
	}
	else if (!strnicmp(val, "off", strlen("off"))) {
		on = 0;
	}
	else {
		return cnt;
	}

	if (!strnicmp(buf, "temp=", strlen("temp="))) {
		led = ENV_LED_TEMPERR;
	}
	else if (!strnicmp(buf, "disk=", strlen("disk="))) {
		led = ENV_LED_DISKERR;
	}
	else if (!strnicmp(buf, "ps=", strlen("ps="))) {
		led = ENV_LED_PSERR;
	}
	else if (!strnicmp(buf, "all=", strlen("all="))) {
		led = ENV_LED_PSERR | ENV_LED_DISKERR | ENV_LED_TEMPERR;
	}
	else {
		printk("%s: unknown led specified\n", __FUNCTION__);
		return cnt;
	}

	env_set_fspleds((on ? led : 0), on ? 0 : led);

	return cnt;
}

#if ENV_DEBUG_RAW /* { */
static ssize_t env_show_raw(struct device *dev, struct device_attribute *attr, char *buf)
{
	int i,j;
	u8 rv;
	size_t ret = 0;
	static const u8 sensors[] = {
		0x9E, /* CPU1-3 TEMP */
		0x9C, /* ENV_FAN_SPD_ADDR */
		0x9A, /* LM75 */
		0x94, /* PS2 */
		0x92, /* PS1 */
		0x90, /* PS0 */
		ENV_FSP_LED_ADDR, /* 0x7C */
		0x78, /* fanfail states */
		0x76, /* PS0 state   */
		0x74, /* PS1 state ? */
		0x72, /* PS2 state   */
		0x70, /* control */
	};

	static const u8 disk_bps[ENV_DISK_BP_MAX] = {
		0x40,	/* ENV_DISK_BP_ADDR_0 */
		0x42,	/* ENV_DISK_BP_ADDR_1 */
		0x44,	/* ENV_DISK_BP_ADDR_2 */
	};

	for (i = 0; i < ARRAY_SIZE(sensors); i++) {
		if ((sensors[i] & 0xF0) == 0x70) {
			ret += sprintf(buf + ret, "0x%02x:    0x%02x\n",
				sensors[i], envctrl_i2c_read_8574(sensors[i]));
		}
		else if ((sensors[i] & 0xf0) == 0x90) {
			if (sensors[i] == ENV_LM75_ADDR) {
				u16 t;
				(void) envctrl_i2c_read_LM75(
					sensors[i], ENV_LIB_LM75_REG_TEMP, &t);
				ret += sprintf(buf + ret,
					"0x%02x: 0x%04x\n",
					sensors[i], t);
				continue;
			}
			for (j = 0; j < 4; j++) {
				(void) envctrl_i2c_read_8591(sensors[i],j,&rv); 
				ret += sprintf(buf + ret, 
					"0x%02x[%i]: 0x%02x\n",
					sensors[i], j, rv); 
			}
		}
	}
	for (i = 0; i < ENV_DISK_BP_MAX; ++i) {
		ret += sprintf(buf + ret, "0x%02x:    0x%02x\n",
			disk_bps[i], envctrl_i2c_read_8574(disk_bps[i]));
	}
	return ret;
}
#endif /* if ENV_DEBUG_RAW */ /* } */

static void env_do_poweroff(void)
{
	static int po_inprog = 0;

	if (po_inprog) {
		return;
	}

	po_inprog = 1;
	printk(KERN_CRIT "%s: WARNING: Powering-off the system now!\n",
		current->comm);

	kernel_power_off();
}

static void env_do_shutdown(void)
{
	static int sd_inprog = 0;
	static char *envp[] = {
		"HOME=/", "TERM=linux", 
		"PATH=/sbin:/usr/sbin:/bin:/usr/bin", NULL };
	char *argv[] = { "/sbin/shutdown", "-h", "now", NULL };
 
	if (sd_inprog != 0)
		return;
 
	sd_inprog = 1;

	printk(KERN_CRIT "%s: WARNING: Shutting down the system now.\n",
		current->comm);

	if (call_usermodehelper("/sbin/shutdown", argv, envp, 0)) {
		printk(KERN_CRIT "%s: WARNING: system shutdown failed!\n",
			current->comm);
		/* unlikely to succeed, but we could try again */
		sd_inprog = 0;  
	}
}

#define WARN_INTERVAL (30 * HZ)
static void env_check_therms(void)
{
	u8 tempbuf;
	u8 rawbuf;
	u8 tempmax_cpu         = 0;
	u8 tempmax_noncpu      = 0;
	u8 ledtemp = 0;

	struct list_head *tmp;
	struct therm     *t;

	list_for_each(tmp, &env_therms) {
		t = list_entry(tmp, struct therm, list);
		if (t->disabled) {
			/* ignore "disabled" thermisters */
			continue;
		}
		tempbuf = env_read_thermister(t, &rawbuf);
		if (t->type == therm_type_cpu) {
#if ENV_DEBUG_TEMP
			if (env_forced_temp)  { 
				tempbuf = env_forced_temp; 
			}
#endif
			if (tempbuf > tempmax_cpu) {
				tempmax_cpu = tempbuf;
			}
		}
		else if (t->type == therm_type_noncpu) { 
			if (tempbuf > tempmax_noncpu) {
				tempmax_noncpu = tempbuf;
			}
		}

		if (t->cfg->shutdown_temp && tempbuf >= t->cfg->shutdown_temp) {
			printk(KERN_CRIT
				"%s: WARNING: %s temperature %u C meets or "
				"exceeds shutdown threshold %u C\n",
				ENV_DEVNAME, t->cfg->name, 
				tempbuf, t->cfg->shutdown_temp); 

			// set LEDs
			spin_lock(&fspled_lock);
			generr_setters++;
			temperr_setters++;
			env_set_fspleds((ENV_LED_GENERR ^ ENV_LED_TEMPERR), 0);
			spin_unlock(&fspled_lock);

			env_set_fanspeed(&env_cpu_fans, ENV_FAN_MAX, 0);
			env_set_fanspeed(&env_psu_fans, ENV_FAN_MAX, 0);

			if (env_poweroff_pfn) {
				env_poweroff_pfn();
			}
			return;
		}
		else if (t->cfg->warn_temp && tempbuf >= t->cfg->warn_temp) {
			if (time_after(jiffies, t->last_warn + WARN_INTERVAL)) {
				t->last_warn = jiffies;
				printk(KERN_CRIT
					"%s: WARNING: %s temperature %u C "
					"meets or exceeds warning threshold "
					"%u C\n",
					ENV_DEVNAME, t->cfg->name,
					tempbuf, t->cfg->warn_temp);
				printk(KERN_CRIT
					"%s: WARNING: system will shut down "
					"at %s temperature %u C\n",
					ENV_DEVNAME, t->cfg->name,
					t->cfg->shutdown_temp);
			}
		}
		else {
			/* Use last_warn as a flag to reduce I2C traffic--
			 * because there is no hysteresis, this could cause 
			 * unlimited warn messages if we hover around warn_temp
			 */
			if (t->last_warn) {
				t->last_warn = 0;
				// clear LEDs
				spin_lock(&fspled_lock);
				generr_setters--;
				temperr_setters--;
				
				if ((generr_setters == 0) && (temperr_setters == 0)) {
					ledtemp = (ENV_LED_GENERR ^ ENV_LED_TEMPERR);
				} else if (temperr_setters == 0) {
					ledtemp = ENV_LED_TEMPERR;
				}
				env_set_fspleds(0, ledtemp);
				
				spin_unlock(&fspled_lock);
			}
		}
	}

	env_set_fanspeed(
		&env_cpu_fans, 
		env_suggest_fanspeed(env_cpu_fanspeed_table, tempmax_cpu), 0);
	env_set_fanspeed(
		&env_psu_fans, 
		env_suggest_fanspeed(env_ps_fanspeed_table, tempmax_noncpu), 0);
}


static int kenvctrld(void *__unused)
{
	int poll_interval = 60 * HZ; /* default value */

	if (envctrl_mon_interval > 0) {
		printk(KERN_INFO "%s: %s monitoring interval: %i seconds\n",
			ENV_DEVNAME, current->comm, envctrl_mon_interval);
		poll_interval = envctrl_mon_interval * HZ;
	}

/*      with this enabled 2.6.32 on e450 OOPS
	daemonize("kenvctrld");
	allow_signal(SIGKILL);
*/	

	printk(KERN_INFO "%s: %s starting\n", ENV_DEVNAME, current->comm);
	while (!kthread_should_stop()) {
		env_check_therms();
		env_ack_fan_wdt();

		current->state = TASK_INTERRUPTIBLE;
		schedule_timeout(poll_interval);
		current->state = TASK_RUNNING;

		if (signal_pending(current)) {
			break;
		}
		
	}
	printk(KERN_INFO "%s: %s exiting\n", ENV_DEVNAME, current->comm);

	return 0;
}

static void env_cleanup(void)
{
	/* XXX: env_cleanup_thermisters(), env_cleanup_fsp(), etc */
	struct therm *therm;
	struct list_head *tmp1, *tmp2;	

	if (env_device) {
		device_remove_file(&(env_device->dev), &dev_attr_fanspeed);
		device_remove_file(&(env_device->dev), &dev_attr_thermisters);
		device_remove_file(&(env_device->dev), &dev_attr_power);
		device_remove_file(&(env_device->dev), &dev_attr_disks);
		device_remove_file(&(env_device->dev), &dev_attr_keyswitch);
		device_remove_file(&(env_device->dev), &dev_attr_fspleds);
#if ENV_DEBUG_TEMP
		device_remove_file(&(env_device->dev), &dev_attr_forced_temp);

		device_remove_file(&(env_device->dev), &dev_attr_lm75);
		device_remove_file(&(env_device->dev), &dev_attr_8583);
#endif
#if ENV_DEBUG_RAW
		device_remove_file(&(env_device->dev), &dev_attr_raw);
#endif
		platform_device_unregister(env_device);
		env_device = NULL;
	}

	if (kenvctrld_task) {
		(void) kthread_stop(kenvctrld_task);
		kenvctrld_task = NULL;
	}

	if (env_fsp_irq) {
		/* disable FSP interrupts */
		envctrl_i2c_write_8574(ENV_FSP_CTRL_ADDR, ENV_FSP_DFLOP_INIT0);
		envctrl_i2c_write_8574(ENV_FSP_CTRL_ADDR, ENV_FSP_DFLOP_INIT1);
		free_irq(env_fsp_irq, (void *) &env_fsp_act_timer);
		env_fsp_irq = 0;
	}

	// we need to extinguish LEDs when we unload, else they stay on even
	// when the system is powered down!
	envctrl_i2c_write_8574(ENV_FSP_LED_ADDR, ENV_FSP_LEDMSK | ENV_FSP_KEYMSK);
	
	/* check if timer has been initialised before we call del_timer_sync() */
	if (env_fsp_act_timer_init) {
		del_timer_sync(&env_fsp_act_timer);
		env_fsp_act_timer_init = 0;
	}

	/* Blast Fans, though watchdog should kick-in as well... */
	env_set_fanspeed(&env_cpu_fans, ENV_FAN_MAX, 0);
	env_set_fanspeed(&env_psu_fans,  ENV_FAN_MAX, 0);
	env_set_fanspeed(&env_afb_fan, ENV_FAN_MAX, 0);

	list_for_each_safe(tmp1, tmp2, &env_therms) {
		therm = list_entry(tmp1, struct therm, list);
		list_del(tmp1);
		kfree(therm);
	}

	if (envctrl_adap) {
		i2c_put_adapter(envctrl_adap);
		envctrl_adap = NULL;
	}

	return;
}

#if ENV_DEBUG_THERMS
static void dump_therms(void)
{
	struct therm *t;
	struct list_head *tmp1, *tmp2;	

	list_for_each_safe(tmp1, tmp2, &env_therms) {
		t = list_entry(tmp1, struct therm, list);
		printk("0x%02x[%u] min %i, warn %i, shut %i, "
			"num %i, den %i, %s\n",
			t->cfg->addr, t->cfg->port, t->cfg->min_temp, 
			t->cfg->warn_temp, t->cfg->shutdown_temp, 
			t->cfg->num, t->cfg->den, t->cfg->name);
	}
}
#endif

static int __init env_init(void)
{
	int err = 0;
	int ret;

	env_parse_overtemp_option(envctrl_overtemp);

	nex = pcf_envctrl_get_nexus();
	if (nex->type != TYPE_ENVCTRL) {
		printk("%s: nexus type %i is not ENVCTRL\n", 
			__FUNCTION__, nex->type);
		return -ENODEV;
	}

	envctrl_adap = i2c_get_adapter(nex->i2c_id);
	if (!envctrl_adap) {
		printk("%s: i2c_get_adapter_id(%i) returned NULL\n",
			__FUNCTION__, nex->i2c_id);
		return -ENODEV;
	}

	env_lib_init(nex, envctrl_adap, ENVLIB_FLAG_8591_NO_AOE);

	if (! env_init_thermisters()) {
		err = -EINVAL;
		goto out_err;
	}

#if ENV_DEBUG_THERMS
	dump_therms();
#endif

	if (! env_init_fan_wdt()) {
		err = -EINVAL;
		goto out_err;
	}

	if (! env_init_fsp()) {
		err = -EINVAL;
		goto out_err;
	}

#if ENV_DEBUG_I2CDEV
	env_lib_scan_bus(0,0,0);
#endif
	kenvctrld_task = kthread_run(kenvctrld, NULL, "kenvctrld");
	if (IS_ERR(kenvctrld_task)) {
		err = PTR_ERR(kenvctrld_task);
		kenvctrld_task = NULL;
		printk(KERN_ERR "%s: unable to create monitor task\n", 
			ENV_DEVNAME);
		goto out_err;
	}

	env_device = platform_device_register_simple(
		ENV_DEVNAME, -1, NULL, 0);
	if (IS_ERR(env_device)) {
		err = PTR_ERR(env_device);
		env_device = NULL;
		printk(KERN_ERR "%s: unable to register sysfs device nodes\n",
			ENV_DEVNAME);
		goto out_err;
	}
	ret = device_create_file(&(env_device->dev), &dev_attr_fanspeed);
	ret = device_create_file(&(env_device->dev), &dev_attr_thermisters);
	ret = device_create_file(&(env_device->dev), &dev_attr_power);
	ret = device_create_file(&(env_device->dev), &dev_attr_disks);
	ret = device_create_file(&(env_device->dev), &dev_attr_keyswitch);
	ret = device_create_file(&(env_device->dev), &dev_attr_fspleds);
#if ENV_DEBUG_TEMP
	ret = device_create_file(&(env_device->dev), &dev_attr_forced_temp);
	ret = device_create_file(&(env_device->dev), &dev_attr_lm75);
	ret = device_create_file(&(env_device->dev), &dev_attr_8583);
#endif
#if ENV_DEBUG_RAW
	ret = device_create_file(&(env_device->dev), &dev_attr_raw);
#endif

	printk(KERN_INFO "%s: driver version %s initialized\n", 
		ENV_DEVNAME, ENV_VERSION);

	/* only returns error if already running, which it isn't... */
	(void) wake_up_process(kenvctrld_task);

	return 0;

out_err:
	env_cleanup();
	return err;
}


module_init(env_init);
module_exit(env_cleanup);
MODULE_LICENSE("GPL");
