/* $Id$
 * env_envctrltwo.c: Environmental monitoring and front status panel
 *                   support for SUNW,envctrltwo (E250)
 *
 * NOTE: Malfunction of this driver may render your system permanently
 *       damaged.  By using this driver you assume all risks due to 
 *       improper operation or otherwise.
 *
 * Copyright (C) 2004-2006 Eric Brower (ebrower@gmail.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* 
 * TODOs: move overtemp handling to library code
 *        implement diagnostic return values for all I2C-access functions
 *        (i2c) i2c-core cannot be used in_interrupt() due to semaphores
 * 	  (i2c) bus-atomicity for sending multiple messages (read,xor,write)
 *	  better IRQ cause reporting/handling (see IDEAs below)
 *	  provide a4 write access?
 *
 * IDEAs: monitor keyswich-- secure == sysctl kernel.stop-a = 0 
 *                           diag   == sysctl kernel.printk increased loglevel
 *	  override sysctl kernel.scons-poweroff if overtemp_action == poweroff
 *	  decrease on-demand bus traffic with sensor periodic polling
 *        periodic polling of ps and disk sensors (and interrupts) to set leds
 * 
 * KNOWN ISSUES: For as-of-yet-unknown reasons, the i2c controller can get
 *	 wedged, leaving the driver without access to devices.  The temporary
 *	 workaround is to reload the i2c_envctrl and envctrltwo drivers.  A
 *	 better solution (in lieu of identifying the exact problem) is to
 * 	 implement PCF8584 reset logic to be triggered during such a condition.
 */

#include <linux/module.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/delay.h>
#include <linux/ioport.h>
#include <linux/init.h>
#include <linux/miscdevice.h>
#include <linux/mm.h>
#include <linux/slab.h>
#include <linux/kernel.h>
#include <linux/i2c.h>
#include <linux/kmod.h>
#include <linux/interrupt.h>
#include <linux/kthread.h>
#include <linux/err.h>
#include <linux/spinlock.h>
#include <linux/reboot.h>
#include <linux/sysfs.h>
#include <linux/platform_device.h>

#include <asm/of_device.h>
#include <asm/uaccess.h>
#include <asm/auxio.h>

#include <asm/i2c-envctrl.h>
#include "env_envctrltwo.h"
#include "env_lib.h"

#define ENV2_DEVNAME	"envctrltwo"
#define ENV2_VERSION    "0.7.0"

#define ENV2_DEBUG_THERMS (0) /* dump thermisters found and values probed  */
#define ENV2_DEBUG_FAN    (0) /* user fanspeed control via sysfs - DANGER! */
#define ENV2_DEBUG_TEMP   (0) /* user CPU temp control via sysfs           */
#define ENV2_DEBUG_RAW    (0) /* ro raw sensor data via sysfs              */
#define ENV2_DEBUG_POWER  (0) /* write control for PS nodes                */
#define ENV2_DEBUG_INTR   (0) /* log all interrupts                        */

static struct pcf_envctrl_nexus *nex;
static struct i2c_adapter *envctrl_adap;

/*
 * OBP "thermisters" property
 * Note: "name" is actually variable length,
 * but null-padded to at least 4 characters
 */
struct env2_thermister_prop {
	u32	addr;
	u32	port;
	u32	min_temp;
	u32	warn_temp;
	u32	shutdown_temp;
	u32	num;
	u32	den;
	u8	name[4];
}__attribute((packed));

typedef enum env2_therm_type {
	therm_type_cpu = 0,
	therm_type_noncpu,
} thermtype_t;

struct therm {
	struct env2_thermister_prop  *cfg;
	thermtype_t                  type; 
	unsigned long                last_warn;
	struct list_head             list;
};

static struct list_head env2_therms = LIST_HEAD_INIT(env2_therms);
static u8 env2_cpu_temp_table[256];
static u8 env2_cpu_fanspeed_table[256];
static u8 *env2_therms_data;

static struct timer_list env2_fsp_act_timer;
static unsigned int env2_fsp_irq;
static u8 env2_pstate = 0xFF;

static struct task_struct *kenvctrld_task;

static void env2_do_poweroff(void);
static void env2_do_shutdown(void);
static void (*env2_poweroff_pfn)(void) = env2_do_shutdown;
static int env2_fanspeed_forced = 0;
static int env2_fsp_act_timer_init = 0;

/* PS States */
typedef enum fru_state {
	fru_state_unknown = 0,
	fru_state_ok,
	fru_state_absent,
	fru_state_absent_attn,
	fru_state_fail,
} fru_state_t;

/* TODO: wasteful-- use single bitmask copy */
static fru_state_t env2_psu_states[ENV2_PS_MAX];
static fru_state_t env2_disk_states[ENV2_DISK_MAX];

static int env2_check_psu_state(void);
static int env2_check_disk_state(void);
static spinlock_t env2_psu_lock = SPIN_LOCK_UNLOCKED;
static spinlock_t env2_disk_lock = SPIN_LOCK_UNLOCKED;

/* sysfs sh!te goes here... */
static struct platform_device *env2_device;

static ssize_t env2_show_fanspeed(struct device *d, struct device_attribute *attr, char *buf);
#if ENV2_DEBUG_FAN
	static ssize_t env2_store_fanspeed(
		struct device *, const char *, size_t);
	static DEVICE_ATTR(fanspeed, \
		S_IRUGO|S_IWUSR, env2_show_fanspeed, env2_store_fanspeed);
#else 
	static DEVICE_ATTR(fanspeed, S_IRUGO, env2_show_fanspeed, NULL);
#endif
#if ENV2_DEBUG_TEMP
	static u8 env2_forced_temp;
	static ssize_t env2_store_temp(struct device *, const char *, size_t);
	static ssize_t env2_show_temp (struct device *, struct device_attribute *attr, char *buf);
	static DEVICE_ATTR(forced_temp, \
		S_IRUGO | S_IWUSR, env2_show_temp, env2_store_temp);
#endif
static ssize_t env2_show_thermisters(struct device *d, struct device_attribute *attr, char *buf);
static DEVICE_ATTR(thermisters, S_IRUGO, env2_show_thermisters, NULL);
#if ENV2_DEBUG_POWER
static ssize_t env2_show_power(struct device *d, struct device_attribute *attr, char *buf);
static ssize_t env2_store_power(struct device *, struct device_attribute *attr, const char *buf, size_t cnt);
static DEVICE_ATTR(power, S_IRUGO|S_IWUSR, env2_show_power, env2_store_power);
#else
static ssize_t env2_show_power(struct device *d, struct device_attribute *attr, char *buf);
static DEVICE_ATTR(power, S_IRUGO, env2_show_power, NULL);
#endif /* ENV2_DEBUG_POWER */
static ssize_t env2_show_disks(struct device *d, struct device_attribute *attr, char *buf);
static ssize_t env2_store_disks(struct device *d, struct device_attribute *attr, const char *buf, size_t cnt);
static DEVICE_ATTR(disks, S_IRUGO|S_IWUSR, env2_show_disks, env2_store_disks);
static ssize_t env2_show_keyswitch(struct device *d, struct device_attribute *attr, char *buf);
static DEVICE_ATTR(keyswitch, S_IRUGO, env2_show_keyswitch, NULL);
static ssize_t env2_show_fspleds(struct device *d, struct device_attribute *attr, char *buf);
/* we can set LEDs, but until we have bus-atomicity it is racy */
static ssize_t env2_store_fspleds(struct device *d, struct device_attribute *attr, const char *b, size_t c);
static DEVICE_ATTR(fspleds, \
	S_IRUGO|S_IWUSR, env2_show_fspleds, env2_store_fspleds);
#if ENV2_DEBUG_RAW
	static ssize_t env2_show_raw (struct device *, struct device_attribute *attr, char *buf);
	static DEVICE_ATTR(raw, S_IRUGO, env2_show_raw, NULL);
	/* a4raw contains security-sensitive RSC configuration data */
	static ssize_t env2_show_a4raw (struct device *, struct device_attribute *attr, char *buf);
	static DEVICE_ATTR(a4raw, S_IRUSR, env2_show_a4raw, NULL);
#endif /* ENV2_DEBUG_RAW */

/* specify "envctrl_overtemp=" to override the default over-temperature
 * action.  Allowable values are poweroff, shutdown and none (dangerous!).
 * The default is shutdown to be idiomatic with existing Linux/SPARC envctrl
 * drivers.
 * They all ought to be "poweroff" for safety-- though this requires
 * tweaking the scons_poweroff sysctl on kernels that support it,
 * otherwise serial console based systems will not poweroff properly.
 */
static int __init env2_parse_overtemp_option(const char *str)
{
	if (NULL == str) {
		return 1;
	}
 
	if (!strncmp(str, "poweroff", strlen("poweroff"))) {
		env2_poweroff_pfn = env2_do_poweroff;
		return 1;
	}
	else if (!strncmp(str, "shutdown", strlen("shutdown"))) {
		env2_poweroff_pfn = env2_do_shutdown;
		return 1;
	}
#if ENV2_DEBUG_TEMP
	else if (!strncmp(str, "none", strlen("none"))) {
		env2_poweroff_pfn = NULL;
		printk(KERN_INFO 
			"%s: WARNING: thermal shutdown/power-off disabled!\n",
			ENV2_DEVNAME);
		return 1;
	}
#endif
	else {
		printk(KERN_ERR 
			"%s: unrecognized envctrl_overtemp value \"%s\" ignored\n", 
			ENV2_DEVNAME, str);
	}
	return 0;
}

static char *envctrl_overtemp;
module_param(envctrl_overtemp, charp, 0);
MODULE_PARM_DESC(envctrl_overtemp,
        "over-temperature action: poweroff or shutdown (default)");

static int envctrl_mon_interval = 60;
module_param(envctrl_mon_interval, uint, 0);
MODULE_PARM_DESC (envctrl_mon_interval,
	"environmental monitoring poll interval (seconds)");

static inline struct property * env_get_property(const char *name, int *lenp)
{
	if (lenp) {
		*lenp = -1;
	}
	return of_find_property(nex->ofdev->node, name, lenp);
}

static char * env2_fru_state_to_str(fru_state_t xlate)
{
		switch(xlate) {
			case (fru_state_ok):
				return "OK";
			case (fru_state_absent):
				return "ABSENT";
			case (fru_state_absent_attn):
				return "ABSENT/ATTN";
			case (fru_state_fail):
				return "FAIL";
			case (fru_state_unknown):
				/* FALL-THRU */
			default:
				break;
		}
		return "UNKNOWN";
}

static u8 env2_get_power(void);
static int env2_check_psu_state(void)
{
	static int init = 0;

	int i;
	int changes = 0;

	u8 state;
	fru_state_t nstate;
	unsigned long flags;

	/* TODO: call from tasklet and get rid of _irqsave */
	spin_lock_irqsave(&env2_psu_lock, flags);
	state = env2_get_power();
	for (i = 0; i < ENV2_PS_MAX; i++) {
		nstate = (ENV2_PS_PRES(state, i) ?
					(ENV2_PS_ERR(state, i) ? 
						fru_state_fail : fru_state_ok) :
							fru_state_absent);
		if (nstate == env2_psu_states[i]) {
			/* no change */
			continue;
		}

		/* don't print state change if we are initializing */
		if (init) {
			printk(KERN_WARNING 
				"%s: PS%i: changed state to %s\n",
				ENV2_DEVNAME, i, env2_fru_state_to_str(nstate));
		}
		env2_psu_states[i] = nstate;
		changes++;
	}
	if (!init) {
		init++;
	}
	spin_unlock_irqrestore(&env2_psu_lock, flags);
	return changes;
}

static u8 env2_get_diskpres(void);
static u8 env2_get_diskstat(void);
static int env2_check_disk_state(void)
{
	static int d_init = 0;

	int i;
	int changes = 0;

	u8 pres, stat;
	fru_state_t nstate;
	unsigned long flags;

	/* TODO: call from tasklet and get rid of _irqsave */
	spin_lock_irqsave(&env2_disk_lock, flags);
	pres = env2_get_diskpres();
	stat = env2_get_diskstat();
	for (i = 0; i < ENV2_DISK_MAX; i++) {

		nstate = (!(pres & (1<<(i))) ? 
				((!(stat & (1<<(i)))) ? 
				 	fru_state_fail : fru_state_ok) : 
					((!(stat & (1<<(i)))) ? 
				 		fru_state_absent_attn : 
						fru_state_absent)); 

		if (nstate == env2_disk_states[i]) {
			/* no change */
			continue;
		}

		/* don't print state change if we are initializing */
		if (d_init) {
			// TODO: don't print absent/attn or fail  ???
			printk(KERN_WARNING 
				"%s: DISK%i: changed state to %s\n",
				ENV2_DEVNAME, i, env2_fru_state_to_str(nstate));
		}
		env2_disk_states[i] = nstate;
		changes++;
	}
	if (!d_init) {
		d_init++;
	}
	spin_unlock_irqrestore(&env2_disk_lock, flags);
	return changes;
}

static int __init env2_init_thermisters(void) 
{
	int i;
	int ntherms = 0;
	struct env2_thermister_prop *t_itr;
	u8 *start, *end;

	struct therm *therm;
	struct list_head *tmp1, *tmp2;	

	struct property *pp;

	/*
	 * The lookup table length is short by two entries that 
	 * represent the coolest thermister raw values 
	 * (indecies 254, 255).  These lower values are reasonably
	 * expected, so we allocate a bigger table and shift the table
	 * values down by two.  We'll duplicate the value at the top of
	 * the table for the two new slots (though these are cooking
	 * temperatures we'd not expect to see).
	 */
	pp = env_get_property("cpu-temp-factors", &i);
	if (i != 254) {
		printk(KERN_ERR "%s: unexpected cpu-temp-factors table size " \
			"(is %u, expected %u)\n", ENV2_DEVNAME, i, 254);
		goto out_err;
	}

	/* TODO: we should use the in-memory copy and index -2 */
	memcpy((u8*)&env2_cpu_temp_table +2, pp->value, i);
	env2_cpu_temp_table[0] = \
		env2_cpu_temp_table[1] = env2_cpu_temp_table[2];

	/* Same drill with cpu-fan-speeds table...
	 *
	 * Observed behavior differs from OBP cpu-fan-speeds table;
	 * it seems that 162 is the minimum speed at which the fan can run,
	 * so don't be confused by the discrepancy between table and readings
	 * at the low end of the scale.
	 */
	pp = env_get_property("cpu-fan-speeds", &i);
	if (i != 254) {
		printk(KERN_ERR "%s: unexpected cpu-fan-speeds table size " \
			"(is %u expected %u)\n", ENV2_DEVNAME, i, 254);
		goto out_err;
	}
	
	/* TODO: we should use the in-memory copy and index -2 */
	memcpy((u8*)&env2_cpu_fanspeed_table +2, pp->value, i);
	env2_cpu_fanspeed_table[0] = \
		env2_cpu_fanspeed_table[1] = env2_cpu_fanspeed_table[2];

	/* TODO: we don't modify this value-- use the OF pointer... */
	pp = env_get_property("thermisters", &i);
	if (!i) {
		printk(KERN_ERR "%s: no thermister data found in OBP\n",
			ENV2_DEVNAME);
		goto out_err;
	}
	env2_therms_data = pp->value;
	if (!env2_therms_data) {
		printk(KERN_ERR "%s: failed to obtain thermister table\n",
			ENV2_DEVNAME);
		goto out_err;
	}
	
	start = env2_therms_data;
	end   = start + i - 1;

	while (start < end &&
	      ((end - start) >= sizeof(struct env2_thermister_prop) - 1)) {
		t_itr = (struct env2_thermister_prop *) start;
		start = t_itr->name + strlen(t_itr->name) + 1;

		therm = kmalloc(sizeof(struct therm), GFP_KERNEL);
		if (!therm) {
			goto out_err;
		}

		/* init one thermister */
		therm->cfg = t_itr;
		therm->last_warn = 0;
		therm->type = 
			(!strncmp(therm->cfg->name, "CPU", strlen("CPU"))) ?
			therm_type_cpu : therm_type_noncpu;
		list_add_tail(&(therm->list), &env2_therms);
		ntherms++;
	}
	return ntherms;

out_err:
	list_for_each_safe(tmp1, tmp2, &env2_therms) {
		therm = list_entry(tmp1, struct therm, list);
		list_del(tmp1);
		kfree(therm);
	}
	return 0;
}

static u8 env2_get_keyswitch(void);
static char* env2_keypos_to_str(u8 kpos);
static irqreturn_t env2_fsp_intr(int irq, void *dev_id)
{
	/* TODO:
	 * known interrupt sources by inverted bit:
	 *   0xBF - PS          (1<<6, 0x40)
	 *   0xDF - Keyswitch   (1<<5, 0x20)
	 *   0xEF - Fan         (1<<4, 0x10)
	 *   0xFB - Disk        (1<<2, 0x04)
	 */

	int known = 0;
	u8 srcs = envctrl_i2c_read_8574(ENV2_FSP_CTRL_ADDR);

	/* The ack seems unnecessary-- reading appears sufficient
	 * envctrl_i2c_write_8574(ENV2_FSP_CTRL_ADDR, 0xFF);
	 */

#if ENV2_DEBUG_INTR
	printk(KERN_DEBUG "%s: %s: interrupt 0x%02x (masked 0x%02x)\n", 
		ENV2_DEVNAME, __FUNCTION__, srcs, srcs ^ env2_pstate);
#endif

	/* 
	 * Handle events -- TODO: defer to tasklet 
	 */
	if ((srcs ^ env2_pstate) & 0x40) {
		env2_check_psu_state();
		srcs |= 0x40;
		known++;
	}
	if ((srcs ^ env2_pstate) & 0x20) {
		u8 kpos = env2_get_keyswitch();
		if (kpos) {
			// check prior state-- occurs with some PSU intrs
			printk(KERN_INFO "%s: keyswitch set to %s\n",
				ENV2_DEVNAME, env2_keypos_to_str(kpos));
			srcs |= 0x20;
		}
		known++;
	}
	if ((srcs ^ env2_pstate) & 0x10) {
		// TODO: handle fan change events
		printk(KERN_INFO "%s: fan status change detected\n",
			ENV2_DEVNAME);
		srcs |= 0x10;
		known++;
	}
	if ((srcs ^ env2_pstate) & 0x04) {
		env2_check_disk_state();
		srcs |= 0x04;
		known++;
	}

	if (!known) {
		printk(KERN_DEBUG "%s: unknown interrupt source (0x%02x)\n", 
			ENV2_DEVNAME, srcs);
	}

	env2_pstate = srcs;
	return IRQ_HANDLED;
}

static void env2_fsp_activity(unsigned long unused)
{
	u8 rd;
	unsigned long flags;

	/* TODO: this must be made i2c-atomic-- not just spinlocked */
	spin_lock_irqsave(&nex->lock, flags);

	__envctrl_i2c_read_8574(ENV2_FSP_LED_ADDR);
	/* always write keyswitch mask-- key bits can be unset, but not set! */
	rd = (__envctrl_i2c_read_8574(ENV2_FSP_LED_ADDR) ^ ENV2_LED_ACTIVE) 
		| ENV2_FSP_KEYMSK;
	__envctrl_i2c_write_8574(ENV2_FSP_LED_ADDR, rd);

	spin_unlock_irqrestore(&nex->lock, flags);

	mod_timer(&env2_fsp_act_timer, jiffies + (HZ/2));
}

static int __init env2_init_fsp(void) 
{

	u8 leds;

	if (nex->ofdev->num_irqs != 2) {
		printk(KERN_ERR "%s: unexpected number of interrupts "
			"(is %i expected %i)\n", 
			ENV2_DEVNAME, nex->ofdev->num_irqs, 2);
		return 0;
	}

	env2_fsp_irq = nex->ofdev->irqs[1];
	if (request_irq(env2_fsp_irq, &env2_fsp_intr,
			IRQF_SHARED, "SUNW,envctrltwo FSP",
			(void *)&env2_fsp_act_timer)) {
		printk(KERN_ERR "%s: unable to register FSP irq\n", 
			ENV2_DEVNAME);
		env2_fsp_irq = 0;
		return 0;
	}

	/* 
	 * FSP GEN_ERR LED is plumbed OR'ed with auxio LED bit.
	 * Extinguish it, otherwise system appears in error
	 */
	auxio_set_led(AUXIO_LED_OFF);

	leds = envctrl_i2c_read_8574(ENV2_FSP_LED_ADDR);
	envctrl_i2c_write_8574(ENV2_FSP_LED_ADDR, leds | ENV2_FSP_LEDMSK);

	init_timer(&env2_fsp_act_timer);
	env2_fsp_act_timer_init = 1;
	env2_fsp_act_timer.expires  = jiffies + (HZ/2);
	env2_fsp_act_timer.function = env2_fsp_activity;
	add_timer(&env2_fsp_act_timer);
	 
	/* Enable FSP interrupts */
	envctrl_i2c_write_8574(ENV2_FSP_CTRL_ADDR, ENV2_FSP_DFLOP_INIT0);
	envctrl_i2c_write_8574(ENV2_FSP_CTRL_ADDR, ENV2_FSP_DFLOP_INIT1);
	envctrl_i2c_write_8574(ENV2_FSP_CTRL_ADDR, ENV2_FSP_DEVINTR_INIT0);
	envctrl_i2c_write_8574(ENV2_FSP_CTRL_ADDR, ENV2_FSP_DEVINTR_INIT1);

	return 1;
}

/* env2_read_thermister: reads a thermister and returns a translated value
 *   if "raw" is non-null, the raw thermister reading will be stored into it
 */
static u8 env2_read_thermister(struct therm *t, u8 *raw)
{
	u8 val;
	int ret;

	switch (t->cfg->addr & 0xF0) {
		case 0x90:
			ret = envctrl_i2c_read_8591(
				t->cfg->addr, t->cfg->port, &val);
			break;
		default:
			/* BUG() */
			return 0;
	}

	if (ret) {
		printk(KERN_ERR "%s: unable to read thermister %s%s\n",
			ENV2_DEVNAME, t->cfg->name,
			(ret == -EINTR)?" (lost arbitration-- transient)":"");
		return 0;
	}
	if (raw) {
		*raw = val;
	}
#if ENV2_DEBUG_THERMS
	printk("%s: %s raw: %u ", __FUNCTION__, t->cfg->name, val);
#endif
	if (t->type == therm_type_cpu) {
		val = env2_cpu_temp_table[val];
	}
	else if (t->cfg->num && t->cfg->den) {
		val = (val * t->cfg->num) / t->cfg->den;
	}
	else {
		/* perhaps raw? */
		printk(KERN_WARNING
			"%s: no thermister translation for %s (val %i)\n",
			ENV2_DEVNAME, t->cfg->name, val);
	}
#if ENV2_DEBUG_THERMS
	printk("cooked: %u\n", val);
#endif
	return val;
}

/* TODO: change to include return value */
static void env2_set_fanspeed(u8 fanspeed, u8 forced)
{
	if (!forced && env2_fanspeed_forced) {
		return;
	}
	envctrl_i2c_write_8591(ENV2_FAN_ADDR, fanspeed);
}

/* TODO: change for meaningful return value */
static u8 env2_get_fanspeed(void)
{
	u8 rv;
	int val;
	/* TODO: raw values read from the fan do not correspond to
 	* the values written (even after suitable settling time).
 	* This is likely due to approximations in the D2A, so we might
 	* as well artificially inflate the read values here...
	* 
	* Ugh-- consider wraparound at high speeds...
 	*/
	if (! envctrl_i2c_read_8591(ENV2_FAN_ADDR, ENV2_FAN_PORT, &rv)) {
		val = rv * 107 / 100;
		if (val > 255 /* UCHAR_MAX */) {
			return 255; /* UCHAR_MAX or FAN_MAX */
		}
		return (u8) val;
	}
	return 0;
}

static ssize_t env2_show_fanspeed(struct device *dev, struct device_attribute *attr, char *buf)
{
	return sprintf(buf, "FAN: [%u]%s\n", 
		env2_get_fanspeed(), env2_fanspeed_forced ? " (forced)" : "");
}

#if ENV2_DEBUG_FAN
static ssize_t env2_store_fanspeed(struct device *dev, const char *buf, size_t count)
{
	u8 val = (u8) simple_strtoul(buf, NULL, 0);
	if (!val) {
		/* writing 0 un-forces fan speed */
		env2_fanspeed_forced = 0;
		return count;
	}

	env2_fanspeed_forced = 1;
	env2_set_fanspeed(val, 1);

	return count;
}
#endif /* ENV2_DEBUG_FAN */

#if ENV2_DEBUG_TEMP
static ssize_t env2_store_temp(struct device *d, const char *buf, size_t count)
{
	u8 val = (u8) simple_strtoul(buf, NULL, 0);
	env2_forced_temp = val;
	return count;
}

static ssize_t env2_show_temp (struct device *d, struct device_attribute *attr, char *buf)
{
	return sprintf(buf, "%i\n", env2_forced_temp);
}
#endif

static ssize_t env2_show_thermisters(struct device *dev, struct device_attribute *attr, char *buf)
{
	ssize_t ret = 0;
	struct therm *t;
	struct list_head *tmp1, *tmp2;	

	list_for_each_safe(tmp1, tmp2, &env2_therms) {
		t = list_entry(tmp1, struct therm, list);
#if ENV2_DEBUG_THERMS
		ret += sprintf(buf + ret, 
			"%s:\t%u C (0x%02x[%u] min %i, warn %i, shut %i, "
			"num %i, den %i, type %i)\n",
			t->cfg->name, env2_read_thermister(t, NULL),
			t->cfg->addr, t->cfg->port, t->cfg->min_temp, 
			t->cfg->warn_temp, t->cfg->shutdown_temp, 
			t->cfg->num, t->cfg->den, t->type);
#else
		ret += sprintf(buf + ret, 
			"%s:\t[%u C] (min %i, warn %i, shut %i)\n",
			t->cfg->name, env2_read_thermister(t, NULL),
			t->cfg->min_temp, t->cfg->warn_temp, 
			t->cfg->shutdown_temp);
#endif
	}
	return ret;
}

static u8 env2_get_power(void)
{
	/* PS removal can leave stale bits present so write power 
	 * register prior to read-- bits can be unset, but not set! 
	 */ 
	(void) envctrl_i2c_write_8574(ENV2_PS_ADDR, 0xFF);
	return envctrl_i2c_read_8574(ENV2_PS_ADDR);
}

#if ENV2_DEBUG_POWER
static ssize_t env2_store_power(struct device *dev, struct device_attribute *attr, const char *buf, size_t cnt)
{
	u8 val;
	val = simple_strtoul(buf, NULL, 0);

	printk(KERN_DEBUG "%s: writing 0x%02x to 0x%02x\n", 
		__FUNCTION__, val, ENV2_PS_ADDR);
	envctrl_i2c_write_8574(ENV2_PS_ADDR, val);

	return cnt;
}
#endif

static ssize_t env2_show_power(struct device *dev, struct device_attribute *attr, char *buf)
{
	ssize_t ret = 0;

	int i;
	for (i = 0; i < ENV2_PS_MAX; i++) {
		ret += sprintf(buf+ret, "PS%i: [%s]\n",
			i, 
			env2_fru_state_to_str(env2_psu_states[i]));
	}
#if ENV2_DEBUG_POWER
	{
		u8 stat = env2_get_power();
		ret += sprintf(buf+ret, "RAW: [0x%02x]\n", stat);
	}
#endif
	return ret;	
}

static u8 env2_get_diskpres(void)
{
	(void)envctrl_i2c_write_8574(ENV2_DISKPRES_ADDR, 0xFF);
	return envctrl_i2c_read_8574(ENV2_DISKPRES_ADDR);
}

static u8 env2_get_diskstat(void)
{
	return envctrl_i2c_read_8574(ENV2_DISKSTAT_ADDR);
}

static ssize_t env2_show_disks(struct device *dev, struct device_attribute *attr, char *buf)
{
	ssize_t ret = 0;
	int i; 
#if 0
	u8 pres = env2_get_diskpres();
	u8 stat = env2_get_diskstat();
	for (i = 0; i < ENV2_DISK_MAX; i++) {
		ret += sprintf(buf+ret, "DISK%i: [%s]\n", i,
			(!(pres & (1<<(i))) ? 
				((!(stat & (1<<(i)))) ? "FAIL" : "OK") : 
				((!(stat & (1<<(i)))) ? "ABSENT/ATTN"  : 
				"ABSENT")));
	}
#else
	for (i = 0; i < ENV2_DISK_MAX; i++) {
		ret += sprintf(buf+ret, 
			"DISK%i: [%s]\n", 
			i, env2_fru_state_to_str(env2_disk_states[i]));
	}
#endif

	return ret;	
}

static ssize_t env2_store_disks(struct device *dev, struct device_attribute *attr, const char *buf, size_t count)
{
	u8 stat;

	/* We don't first check the disk is present-- we could, though */
	u8 diskno = (u8) simple_strtoul(buf, NULL, 0);
	if (diskno < ENV2_DISK_MAX) {
		char *val = strchr(buf, '=');
		if (val) {
			val++;
			if (!strnicmp(val, "fail", strlen("fail"))) {
				/* TODO: this must be made bus-atomic */
				spin_lock(&nex->lock);
				stat = 
				  __envctrl_i2c_read_8574(ENV2_DISKSTAT_ADDR);
				stat &= ~(1 << (diskno));
				__envctrl_i2c_write_8574(
					ENV2_DISKSTAT_ADDR, stat);
				spin_unlock(&nex->lock);
			}
			else if (!strnicmp(val, "ok", strlen("ok"))) {
				/* TODO: this must be made bus-atomic */
				spin_lock(&nex->lock);
				stat =
				  __envctrl_i2c_read_8574(ENV2_DISKSTAT_ADDR);
				stat |= (1 << (diskno));
				__envctrl_i2c_write_8574(
					ENV2_DISKSTAT_ADDR, stat);
				spin_unlock(&nex->lock);
			}
		}
	}
	env2_check_disk_state(); // TODO: only care about insertion/removal?
	return count;
}

static u8 env2_get_keyswitch(void)
{
	return envctrl_i2c_read_8574(ENV2_FSP_LED_ADDR) & ENV2_FSP_KEYMSK;
}

static char* env2_keypos_to_str(u8 kpos)
{
	switch (kpos) {
		case ENV2_FSP_KEY_OFF: 
			return("OFF"); 
		case ENV2_FSP_KEY_DIAG:
			return("DIAG");
		case ENV2_FSP_KEY_ON:
			return("ON");
		case ENV2_FSP_KEY_LOCK:
			return("LOCKED");
	}
	return ("UKNOWN");
}

static ssize_t env2_show_keyswitch(struct device *dev, struct device_attribute *attr, char *buf)
{
	ssize_t ret = 0;

	ret += sprintf(buf + ret, "KEYSWITCH: [%s]\n", 
		env2_keypos_to_str(env2_get_keyswitch()));

	return ret;	
}

static u8 env2_get_fspleds(void)
{
	return envctrl_i2c_read_8574(ENV2_FSP_LED_ADDR) & ENV2_FSP_LEDMSK;
}

/* TODO: diagnostic return value */
static void env2_set_fspleds(u8 leds_on, u8 leds_off)
{
	u8 rd;
	unsigned long flags;

	/* don't stomp on keyswitch bits */
	if ((leds_on | leds_off) & ENV2_FSP_KEYMSK) {
		return;
	}

	/* TODO: this must be made i2c-atomic-- not just spinlocked */
	spin_lock_irqsave(&nex->lock, flags);

	__envctrl_i2c_read_8574(ENV2_FSP_LED_ADDR);
	/* always write keyswitch mask-- key bits can be unset, but not set! */
	rd = (__envctrl_i2c_read_8574(ENV2_FSP_LED_ADDR) | 
		ENV2_FSP_KEYMSK | leds_off) & ~leds_on;
	__envctrl_i2c_write_8574(ENV2_FSP_LED_ADDR, rd);

	spin_unlock_irqrestore(&nex->lock, flags);
}

static ssize_t env2_show_fspleds(struct device *dev, struct device_attribute *attr, char *buf)
{
	ssize_t ret = 0;
	u8 stat = env2_get_fspleds();

	ret += sprintf(buf + ret, "POWER:      [%s]\n", "ON"); 
	ret += sprintf(buf + ret, "GEN ERROR:  [%s]\n", 
		stat & ENV2_LED_GENERR ? "OFF" : "ON");
	ret += sprintf(buf + ret, "ACTIVITY:   [%s]\n", 
		stat & ENV2_LED_ACTIVE ? "OFF" : "ON");
	ret += sprintf(buf + ret, "DISK ERROR: [%s]\n", 
		stat & ENV2_LED_DISKERR ? "OFF" : "ON");
	ret += sprintf(buf + ret, "TEMP ERROR: [%s]\n", 
		stat & ENV2_LED_TEMPERR ? "OFF" : "ON");
	ret += sprintf(buf + ret, "PS ERROR:   [%s]\n", 
		stat & ENV2_LED_PSERR ? "OFF" : "ON");

	return ret;	
}

static ssize_t env2_store_fspleds(struct device *dev, struct device_attribute *attr, const char *buf, size_t cnt)
{
	char on = -1;
	char led= -1;

	char *val = strchr(buf, '=');
	if (!val) {
		return cnt;
	}
	val++;
	
	if (!strnicmp(val, "on", strlen("on"))) {
		on = 1;
	}
	else if (!strnicmp(val, "off", strlen("off"))) {
		on = 0;
	}
	else {
		return cnt;
	}

	if (!strnicmp(buf, "temp=", strlen("temp="))) {
		led = ENV2_LED_TEMPERR;
	}
	else if (!strnicmp(buf, "disk=", strlen("disk="))) {
		led = ENV2_LED_DISKERR;
	}
	else if (!strnicmp(buf, "ps=", strlen("ps="))) {
		led = ENV2_LED_PSERR;
	}
	else if (!strnicmp(buf, "all=", strlen("all="))) {
		led = ENV2_LED_PSERR | ENV2_LED_DISKERR | ENV2_LED_TEMPERR;
	}
	else {
		printk("%s: unknown led specified\n", __FUNCTION__);
		return cnt;
	}

	env2_set_fspleds((on ? led : 0), on ? 0 : led);

	return cnt;
}

#if ENV2_DEBUG_RAW /* { */
static ssize_t env2_show_raw(struct device *dev, struct device_attribute *attr, char *buf)
{
	int i,j;
	u8 rv;
	size_t ret = 0;
	static const u8 sensors[] = {
		ENV2_FSP_CTRL_ADDR,
		ENV2_PS_ADDR,
		ENV2_DISKPRES_ADDR,
		ENV2_FSP_LED_ADDR,
		ENV2_DISKSTAT_ADDR,
		ENV2_FAN_ADDR,      /* SCSI, PDB, FAN */
		0x09E,              /* CPU therms     */
	};

	for (i = 0; i < ARRAY_SIZE(sensors); i++) {
		if ((sensors[i] & 0xF0) == 0x70) {
			ret += sprintf(buf + ret, "0x%02x:    0x%02x\n",
				sensors[i], envctrl_i2c_read_8574(sensors[i]));
		}
		else if ((sensors[i] & 0xf0) == 0x90) {
			for (j = 0; j < 4; j++) {
				(void) envctrl_i2c_read_8591(sensors[i],j,&rv); 
				ret += sprintf(buf + ret, 
					"0x%02x[%i]: 0x%02x\n",
					sensors[i], j, rv); 
			}
		}
	}
	return ret;
}

static ssize_t env2_show_a4raw(struct device *dev, struct device_attribute *attr, char *buf)
{
	int i;
	u8 mybuf;
	size_t ret = 0;

	for (i = 0; i < 256; i++) {
		if (! envctrl_i2c_read_8570(0xA4, &mybuf, i)) {
			ret += sprintf(buf + ret, "%c", mybuf);
		}
		else {
			break;
		}
	}

	return ret;
}
#endif /* if ENV2_DEBUG_RAW */ /* } */

static void env2_do_poweroff(void)
{
	static int po_inprog = 0;

	if (po_inprog) {
		return;
	}

	po_inprog = 1;
	printk(KERN_CRIT "%s: WARNING: Powering-off the system now!\n",
		current->comm);

	kernel_power_off();
}

static void env2_do_shutdown(void)
{
	static int sd_inprog = 0;
	static char *envp[] = {
		"HOME=/", "TERM=linux", 
		"PATH=/sbin:/usr/sbin:/bin:/usr/bin", NULL };
	char *argv[] = { "/sbin/shutdown", "-h", "now", NULL };
 
	if (sd_inprog != 0)
		return;
 
	sd_inprog = 1;

	printk(KERN_CRIT "%s: WARNING: Shutting down the system now.\n",
		current->comm);

	if (call_usermodehelper("/sbin/shutdown", argv, envp, 0)) {
		printk(KERN_CRIT "%s: WARNING: system shutdown failed!\n",
			current->comm);
		/* unlikely to succeed, but we could try again */
		sd_inprog = 0;  
	}
}

#define WARN_INTERVAL (30 * HZ)
static void env2_check_therms(void)
{
	u8 tempbuf;
	u8 tempmax = 0;
	u8 rawbuf;
	u8 tempmax_raw  = 0;

	struct list_head *tmp;
	struct therm     *t;

	list_for_each(tmp, &env2_therms) {
		t = list_entry(tmp, struct therm, list);
		tempbuf = env2_read_thermister(t, &rawbuf);
		if (t->type == therm_type_cpu) {
#if ENV2_DEBUG_TEMP
			if (env2_forced_temp)  { 
				tempbuf = env2_forced_temp; 
			}
#endif
			if (tempbuf > tempmax) {
				tempmax      = tempbuf;
				tempmax_raw  = rawbuf;
			}
		}
		if (t->cfg->shutdown_temp && tempbuf >= t->cfg->shutdown_temp) {
			printk(KERN_CRIT
				"%s: WARNING: %s temperature %u C meets or "
				"exceeds shutdown threshold %u C\n",
				ENV2_DEVNAME, t->cfg->name, 
				tempbuf, t->cfg->shutdown_temp); 

			env2_set_fspleds(ENV2_LED_GENERR, 0);
			env2_set_fanspeed(ENV2_FAN_MAX, 0);

			if (env2_poweroff_pfn) {
				env2_poweroff_pfn();
			}
			return;
		}
		else if (t->cfg->warn_temp && tempbuf >= t->cfg->warn_temp) {
			/* TODO: set thermal warning LED or gen_err */
			if (time_after(jiffies, t->last_warn + WARN_INTERVAL)) {
				t->last_warn = jiffies;
				printk(KERN_CRIT
					"%s: WARNING: %s temperature %u C "
					"meets or exceeds warning threshold "
					"%u C\n",
					ENV2_DEVNAME, t->cfg->name,
					tempbuf, t->cfg->warn_temp);
				printk(KERN_CRIT
					"%s: WARNING: system will shut down "
					"at %s temperature %u C\n",
					ENV2_DEVNAME, t->cfg->name,
					t->cfg->shutdown_temp);
			}
		}
		else {
			/* Use last_warn as a flag to reduce I2C traffic--
			 * because there is no hysteresis, this could cause 
			 * unlimited warn messages if we hover around warn_temp
			 */
			if (t->last_warn) {
				t->last_warn = 0;
				env2_set_fspleds(0, ENV2_LED_GENERR);
			}
		}
	}
	env2_set_fanspeed(env2_cpu_fanspeed_table[tempmax_raw], 0);
}

static int kenvctrld(void *__unused)
{
	int poll_interval = 5 * HZ;
	if (envctrl_mon_interval > 0) {
		poll_interval = envctrl_mon_interval * HZ;
		printk(KERN_INFO 
			"%s: %s monitoring interval %i seconds\n", 
			ENV2_DEVNAME, current->comm, envctrl_mon_interval);
	}

	daemonize("kenvctrld");
	allow_signal(SIGKILL);

	printk(KERN_INFO "%s: %s starting\n", ENV2_DEVNAME, current->comm);
	while (!kthread_should_stop()) {

		env2_check_therms();
		env2_check_psu_state();
		env2_check_disk_state();

		current->state = TASK_INTERRUPTIBLE;
		schedule_timeout(poll_interval);
		current->state = TASK_RUNNING;

		if (signal_pending(current)) {
			break;
		}
	}
	printk(KERN_INFO "%s: %s exiting\n", ENV2_DEVNAME, current->comm);

	return 0;
}

static void env2_cleanup(void)
{
	/* XXX: env2_cleanup_thermisters(), env2_cleanup_fsp(), etc */
	struct therm *therm;
	struct list_head *tmp1, *tmp2;	

	if (env2_device) {
		device_remove_file(&(env2_device->dev), &dev_attr_fanspeed);
		device_remove_file(&(env2_device->dev), &dev_attr_thermisters);
		device_remove_file(&(env2_device->dev), &dev_attr_power);
		device_remove_file(&(env2_device->dev), &dev_attr_disks);
		device_remove_file(&(env2_device->dev), &dev_attr_keyswitch);
		device_remove_file(&(env2_device->dev), &dev_attr_fspleds);
#if ENV2_DEBUG_TEMP
		device_remove_file(&(env2_device->dev), &dev_attr_forced_temp);
#endif
#if ENV2_DEBUG_RAW
		device_remove_file(&(env2_device->dev), &dev_attr_raw);
		device_remove_file(&(env2_device->dev), &dev_attr_a4raw);
#endif
		platform_device_unregister(env2_device);
		env2_device = NULL;
	}

	if (kenvctrld_task) {
		(void) kthread_stop(kenvctrld_task);
		kenvctrld_task = NULL;
	}

	if (env2_fsp_irq) {
		/* disable FSP interrupts */
		envctrl_i2c_write_8574(
			ENV2_FSP_CTRL_ADDR, ENV2_FSP_INTR_LATCH_CLR);
		free_irq(env2_fsp_irq, (void *) &env2_fsp_act_timer);
		env2_fsp_irq = 0;
	}
	
	/* check if timer has been initialised before we call del_timer_sync() */
	if (env2_fsp_act_timer_init) {
		del_timer_sync(&env2_fsp_act_timer);
		env2_fsp_act_timer_init = 0;
	}

	/* Blast fans */
	env2_set_fanspeed(ENV2_FAN_MAX, 0);

	list_for_each_safe(tmp1, tmp2, &env2_therms) {
		therm = list_entry(tmp1, struct therm, list);
		list_del(tmp1);
		kfree(therm);
	}

	if (envctrl_adap) {
		i2c_put_adapter(envctrl_adap);
		envctrl_adap = NULL;
	}

	return;
}

#if ENV2_DEBUG_THERMS
static void dump_therms(void)
{
	struct therm *t;
	struct list_head *tmp1, *tmp2;	

	list_for_each_safe(tmp1, tmp2, &env2_therms) {
		t = list_entry(tmp1, struct therm, list);
		printk("0x%02x[%u] min %i, warn %i, shut %i, "
			"num %i, den %i, %s\n",
			t->cfg->addr, t->cfg->port, t->cfg->min_temp, 
			t->cfg->warn_temp, t->cfg->shutdown_temp, 
			t->cfg->num, t->cfg->den, t->cfg->name);
	}
}
#endif

static int __init env2_init(void)
{
	int err = 0;
	int ret;

	env2_parse_overtemp_option(envctrl_overtemp);

	nex = pcf_envctrl_get_nexus();
	if (nex->type != TYPE_ENVCTRLTWO) {
		printk("%s: nexus type %i is not ENVCTRLTWO\n", 
			__FUNCTION__, nex->type);
		return -ENODEV;
	}

	envctrl_adap = i2c_get_adapter(nex->i2c_id);
	if (!envctrl_adap) {
		printk("%s: i2c_get_adapter_id(%i) returned NULL\n",
			__FUNCTION__, nex->i2c_id);
		return -ENODEV;
	}

	env_lib_init(nex, envctrl_adap, 0);

//env_lib_scan_bus(0x00, 0x69, 0);
//env_lib_scan_bus(0x71, 0xFF, 0);

	if (! env2_init_thermisters()) {
		err = -EINVAL;
		goto out_err;
	}

#if ENV2_DEBUG_THERMS
	dump_therms();
#endif

	if (! env2_init_fsp()) {
		err = -EINVAL;
		goto out_err;
	}

	kenvctrld_task = kthread_create(kenvctrld, NULL, "kenvctrld");
	if (IS_ERR(kenvctrld_task)) {
		err = PTR_ERR(kenvctrld_task);
		kenvctrld_task = NULL;
		goto out_err;
	}

	env2_device = platform_device_register_simple(
		ENV2_DEVNAME, -1, NULL, 0);
	if (IS_ERR(env2_device)) {
		err = PTR_ERR(env2_device);
		env2_device = NULL;
		goto out_err;
	}
	ret = device_create_file(&(env2_device->dev), &dev_attr_fanspeed);
	ret = device_create_file(&(env2_device->dev), &dev_attr_thermisters);
	ret = device_create_file(&(env2_device->dev), &dev_attr_power);
	ret = device_create_file(&(env2_device->dev), &dev_attr_disks);
	ret = device_create_file(&(env2_device->dev), &dev_attr_keyswitch);
	ret = device_create_file(&(env2_device->dev), &dev_attr_fspleds);
#if ENV2_DEBUG_TEMP
	ret = device_create_file(&(env2_device->dev), &dev_attr_forced_temp);
#endif
#if ENV2_DEBUG_RAW
	ret = device_create_file(&(env2_device->dev), &dev_attr_raw);
	ret = device_create_file(&(env2_device->dev), &dev_attr_a4raw);
#endif

	printk(KERN_INFO "%s: driver version %s initialized\n", 
		ENV2_DEVNAME, ENV2_VERSION);

	/* only returns error if already running, which it isn't... */
	(void) wake_up_process(kenvctrld_task);

	return 0;

out_err:
	env2_cleanup();
	return err;
}


module_init(env2_init);
module_exit(env2_cleanup);
MODULE_LICENSE("GPL");
