/* $Id$
 * env_lib.c: Environmental monitoring library code for 
 *            accessing I2C clients (chips) via i2c-envctrl 
 *            PCF8584 bus nexus interface 
 *
 * Copyright (C) 2004-2005 Eric Brower (ebrower@gmail.com)
 * 
 * TODO: functions should all return explicit error codes
 */

#include <linux/spinlock.h>
#include <linux/limits.h>
#include <asm/i2c-envctrl.h>
#include "env_lib.h"

/* initialized from env_lib_init() routine
 */
static struct i2c_adapter *envctrl_adap;
static spinlock_t *envctrl_lock;
static envlib_flags_t envlib_flags;

/* Initialize library-- 
 * This MUST be done prior to any other library use
 * TODO: make code completely unreadable with BUG() 
 *       calls when init has not been done...
 */ 
void env_lib_init(struct pcf_envctrl_nexus *nex, 
	          struct i2c_adapter *adap,
                  envlib_flags_t flags)
{
	envctrl_adap = adap;
	envctrl_lock = &nex->lock;
	envlib_flags = flags;
}

/* returns 0 on success, error code on failure */
int __env2_i2c_readb(u8 addr, u8 *rv)
{ 
	int ret;
	struct i2c_msg msg = {
		.addr  = (addr>>1),
		.flags = I2C_M_RD,
		.len   = 1,
		.buf   = rv,
	};

	ret = i2c_transfer(envctrl_adap, &msg, 1);
	if (ret == 1) {
		ret = 0;
	}
	return (ret == 1) ? 0 : ret;
}

int env2_i2c_readb(u8 addr, u8 *rv)
{ 
	int ret;
	unsigned long flags;
	spin_lock_irqsave(envctrl_lock, flags);

	ret = __env2_i2c_readb(addr, rv);
	spin_unlock_irqrestore(envctrl_lock, flags);
	return ret;
}

int __env2_i2c_writeb(u8 addr, u8 val)
{ 
	u8 valbuf = val;
	int ret;

	struct i2c_msg msg = {
		.addr  = (addr>>1),
		.flags = 0,
		.len   = 1,
		.buf   = &valbuf,
	};

	if (!envctrl_adap) {
		BUG();
	}

	ret = i2c_transfer(envctrl_adap, &msg, 1);
	if (ret == 1) {
		ret = 0;
	}
	return ret;
}

int env2_i2c_writeb(u8 addr, u8 val)
{ 
	int ret;
	unsigned long flags;

	spin_lock_irqsave(envctrl_lock, flags);
	ret = __env2_i2c_writeb(addr, val);
	spin_unlock_irqrestore(envctrl_lock, flags);
	return ret;
}

u8 __envctrl_i2c_read_8574(u8 addr) 
{
	u8 buf;
	(void) __env2_i2c_readb(addr, &buf);
	return buf;
}
u8 envctrl_i2c_read_8574(u8 addr) 
{
	u8 buf;
	(void) env2_i2c_readb(addr, &buf);
	return buf;
}

int __envctrl_i2c_read_8591(u8 addr, u8 port, u8 *rv)
{
#define PCF8591_AOE (0x40)
	int ret;
	u8 tmp = port | 
	         ((envlib_flags & ENVLIB_FLAG_8591_NO_AOE) ? 0 : PCF8591_AOE);

	struct i2c_msg msg[] = {
		{
			.addr  = (addr>>1),
			.flags = 0,
			.len   = 1,
			.buf   = &tmp,
		},
		{
			/* dummy read-- result of previous conversion cycle */
			.addr  = (addr>>1),
			.flags = I2C_M_RD,
			.len   = 1,
			.buf   = &tmp,
		},
		{
			.addr  = (addr>>1),
			.flags = I2C_M_RD,
			.len   = 1,
			.buf   = rv,
		},
	
	};

	int msgcnt = sizeof(msg) / sizeof(struct i2c_msg);
	if (!envctrl_adap) {
		BUG();
	}
	ret = i2c_transfer(envctrl_adap, msg, msgcnt);
	if (ret == msgcnt) {
		ret = 0;
	}
	return ret;
}

int envctrl_i2c_read_8591(u8 addr, u8 port, u8 *rv)
{
	int ret;
	unsigned long flags;

	spin_lock_irqsave(envctrl_lock, flags);
	ret = __envctrl_i2c_read_8591(addr, port, rv);
	spin_unlock_irqrestore(envctrl_lock, flags);
	return ret;
}

u8 __envctrl_i2c_write_8591(u8 addr, u8 data)
{
	int ret;
	u8 tmp[] = { PCF8591_AOE, data };

	struct i2c_msg msg[] = {
		{
			.addr  = (addr>>1),
			.flags = 0,
			.len   = 2,
			.buf   = tmp,
		}
	};

	int msgcnt = sizeof(msg) / sizeof(struct i2c_msg);

	if (!envctrl_adap) {
		BUG();
	}
	ret = i2c_transfer(envctrl_adap, msg, msgcnt);
	if (ret == msgcnt) {
		ret = 0;
	}
	return ret;
}

u8 envctrl_i2c_write_8591(u8 addr, u8 data)
{
	u8 ret;
	unsigned long flags;

	spin_lock_irqsave(envctrl_lock, flags);
	ret = __envctrl_i2c_write_8591(addr, data);
	spin_unlock_irqrestore(envctrl_lock, flags);
	return ret;
}

u8 __envctrl_i2c_read_8570(u8 addr, u8 *buf, u8 byteno)
{
	int ret;
	int msgcnt;
	struct i2c_msg msg[2];

	if (! buf) {
		return 1;
	}

	msg[0].addr = (addr>>1);
	msg[0].flags = 0;
	msg[0].len = 1;
	msg[0].buf = &byteno;
	msg[1].addr = (addr>>1);
	msg[1].flags = I2C_M_RD;
	msg[1].len = 1;
	msg[1].buf = buf;

	msgcnt = sizeof(msg) / sizeof(struct i2c_msg);

	if (!envctrl_adap) {
		BUG();
	}
	ret = i2c_transfer(envctrl_adap, msg, msgcnt);
	if (ret == msgcnt) {
		ret = 0;
	}
	return ret;
}

u8 envctrl_i2c_read_8570(u8 addr, u8 *buf, u8 byteno)
{
	u8 ret;
	unsigned long flags;

	spin_lock_irqsave(envctrl_lock, flags);
	ret = __envctrl_i2c_read_8570(addr, buf, byteno);
	spin_unlock_irqrestore(envctrl_lock, flags);
	return ret;
}

/* This function is useless-- the TDA8444(T) cannot be read!
 */
int __envctrl_i2c_read_8444(u8 addr, u8 port, u8 *rv)
{
	int ret;
	int msgcnt;
	struct i2c_msg msg[2];
	
	if (! rv) {
		return 1;
	}

	if (port > ENVLIB_TDA8444_PORT_MAX) {
		return 1;
	}

	msg[0].addr = (addr>>1);
	msg[0].flags = 0;
	msg[0].len = 1;
	msg[0].buf = &port;
	msg[1].addr = (addr>>1);
	msg[1].flags = I2C_M_RD;
	msg[1].len = 1;
	msg[1].buf = rv;

	msgcnt = sizeof(msg) / sizeof(struct i2c_msg);

	if (!envctrl_adap) {
		BUG();
	}
	ret = i2c_transfer(envctrl_adap, msg, msgcnt);
	if (ret == msgcnt) {
		ret = 0;
	}
	*rv &= ENVLIB_TDA8444_DATAMSK; /* truncate to 6-bit value */
	return ret;
}

int envctrl_i2c_read_8444(u8 addr, u8 port, u8 *rv)
{
	u8 ret;
	unsigned long flags;

	spin_lock_irqsave(envctrl_lock, flags);
	ret = __envctrl_i2c_read_8444(addr, port, rv);
	spin_unlock_irqrestore(envctrl_lock, flags);
	return ret;
}

int __envctrl_i2c_write_8444(u8 addr, u8 port, u8 data)
{
	int ret;
	u8 buf[2];
	int msgcnt;
	struct i2c_msg msg[1];

	if (port > ENVLIB_TDA8444_PORT_MAX) {
		return -EINVAL;
	}

	buf[0] = port | 0xF0; /* Instruction bits == don't auto-increment */
	buf[1] = data;

	msg[0].addr = (addr>>1);
	msg[0].flags = 0;
	msg[0].len = sizeof(buf);
	msg[0].buf = buf;

	msgcnt = sizeof(msg) / sizeof(struct i2c_msg);

	if (!envctrl_adap) {
		BUG();
	}
	ret = i2c_transfer(envctrl_adap, msg, msgcnt);
	if (ret == msgcnt) {
		ret = 0;
	}
	return ret;
}

int envctrl_i2c_write_8444(u8 addr, u8 port, u8 data)
{
	u8 ret;
	unsigned long flags;

	spin_lock_irqsave(envctrl_lock, flags);
	ret = __envctrl_i2c_write_8444(addr, port, data);
	spin_unlock_irqrestore(envctrl_lock, flags);
	return ret;
}

/* TODO: fix me!!! 
 * The interrupt control device (on SUNW,envctrltwo @ 0x70)
 * doesn't seem to like being read until initialized-- I think...
 */
void env_lib_scan_bus(u8 start, u8 stop, int read)
{
	int rc, i;
	unsigned long flags;
	u8 tmp;
                                                                        
	if (!envctrl_adap) {
		BUG();
	}

	start &= ~0x01; /* mask "read" address bit */
	if (start == 0x00 && stop == 0x00) {
		stop = 0xFF;
	}
                                                                                
	printk(KERN_INFO "%s: scanning bus %s.\n",
		__FUNCTION__, envctrl_adap->name);
	spin_lock_irqsave(envctrl_lock, flags);
	for (i = start; i < stop; i+=2) {
		struct i2c_msg msg[] = {
			{
				.addr  = (i>>1),
				.flags = (read ? I2C_M_RD : 0),
				.len   = 1,
				.buf   = &tmp,
			},
		};

		rc = i2c_transfer(envctrl_adap, msg, 1);
		if (rc == 1) {
			printk("(%02x)",i);
		} else
			printk(".");
	}
	spin_unlock_irqrestore(envctrl_lock, flags);
	printk("\n");
}


/*
 * LM75 support
 */
int __envctrl_i2c_read_LM75(u8 addr, u8 reg, u16 *rv)
{
	int ret;
	u8  buf[2];
	int msgcnt;
	struct i2c_msg msg[2];

	if (! rv) {
		return 1;
	}

	if (reg & ~ENV_LIB_LM75_REG_MSK) {
		return -EINVAL;
	}

	msg[0].addr = (addr>>1);
	msg[0].flags = 0;
	msg[0].len = 1;
	msg[0].buf = &reg;
	msg[1].addr = (addr>>1);
	msg[1].flags = I2C_M_RD;
	msg[1].len = (reg == ENV_LIB_LM75_REG_CFG ? 1 : 2);
	msg[1].buf = buf;

	msgcnt = sizeof(msg) / sizeof(struct i2c_msg);

	if (!envctrl_adap) {
		BUG();
	}
	ret = i2c_transfer(envctrl_adap, msg, msgcnt);
	if (ret == msgcnt) {
		ret = 0;
	}

	/* all registers return MSB/LSB, except CFG */
	if (reg == ENV_LIB_LM75_REG_CFG)
	{
		*rv = buf[0];
	}
	else {
		/* first byte is MSB, second is LSB */
		*rv = buf[0] << 8 | buf[1];
	}
	return ret;
}

int envctrl_i2c_read_LM75(u8 addr, u8 port, u16 *rv)
{
	u8 ret;
	unsigned long flags;

	spin_lock_irqsave(envctrl_lock, flags);
	ret = __envctrl_i2c_read_LM75(addr, port, rv);
	spin_unlock_irqrestore(envctrl_lock, flags);
	return ret;
}

/* Convenience function for reading LM75 temperature
 *
 * NOTE: for simplicity and to remain idomatic with exiting sensors,
 * we truncate the 9-bit signed value to 8 bits unsigned and
 * treat all temperatures < 0 as zero.  We round-up the 0.5 
 * degree bit if present. The effective range reported by this 
 * function is 0-128 degrees C rather than the LM75 supported
 * range of -127.5 C to +127.5 C. That's OK-- we don't threshold
 * for values < 0.
 */
int envctrl_i2c_read_LM75_temp(u8 addr, u8 *rv)
{
	u16 val;
	int ret = envctrl_i2c_read_LM75(addr, ENV_LIB_LM75_REG_TEMP, &val);
	*rv = val >> 8;
	if (*rv > 127 /* SCHAR_MAX */) {
		*rv = 0;
	}
	else {
		*rv += (val & 0x0080) ? 1 : 0;
	}
	return ret;
}

int __envctrl_i2c_write_LM75(u8 addr, u8 reg, u16 data)
{
	int ret;
	u8  buf[3];
	int msgcnt;
	struct i2c_msg msg[1];

	if (reg & ~ENV_LIB_LM75_REG_MSK) {
		return -EINVAL;
	}

	buf[0] = reg;

	if (reg == ENV_LIB_LM75_REG_CFG) {
		buf[1] = data & 0xFF;
		buf[2] = 0;
	}
	else {
		buf[1] = data >> 8;
		buf[2] = data & 0xFF;
	}

	msg[0].addr = (addr>>1);
	msg[0].flags = 0;
	msg[0].len = (reg == ENV_LIB_LM75_REG_CFG ? 2 : 3);
	msg[0].buf = buf;

	msgcnt = sizeof(msg) / sizeof(struct i2c_msg);

	if (!envctrl_adap) {
		BUG();
	}
	ret = i2c_transfer(envctrl_adap, msg, msgcnt);
	if (ret == msgcnt) {
		ret = 0;
	}
	return ret;
}

int envctrl_i2c_write_LM75(u8 addr, u8 reg, u16 data)
{
	u8 ret;
	unsigned long flags;

	spin_lock_irqsave(envctrl_lock, flags);
	ret = __envctrl_i2c_write_LM75(addr, reg, data);
	spin_unlock_irqrestore(envctrl_lock, flags);
	return ret;
}

/*
 * PCF8583 support
 */
int __envctrl_i2c_read_pcf8583(u8 addr, u8 reg, u8 *rv)
{
	int ret;
	int msgcnt;
	struct i2c_msg msg[2];
	
	if (! rv) {
		return -EINVAL;
	}

	if (reg & ~ENV_LIB_8583_REG_MSK) {
		return -EINVAL;
	}

	msg[0].addr = (addr>>1);
	msg[0].flags = 0;
	msg[0].len = 1;
	msg[0].buf = &reg;
	msg[1].addr = (addr>>1);
	msg[1].flags = I2C_M_RD;
	msg[1].len = 1;
	msg[1].buf = rv;

	msgcnt = sizeof(msg) / sizeof(struct i2c_msg);

	if (!envctrl_adap) {
		BUG();
	}
	ret = i2c_transfer(envctrl_adap, msg, msgcnt);
	if (ret == msgcnt) {
		ret = 0;
	}

	return ret;
}

int envctrl_i2c_read_pcf8583(u8 addr, u8 port, u8 *rv)
{
	u8 ret;
	unsigned long flags;

	spin_lock_irqsave(envctrl_lock, flags);
	ret = __envctrl_i2c_read_pcf8583(addr, port, rv);
	spin_unlock_irqrestore(envctrl_lock, flags);
	return ret;
}

int __envctrl_i2c_write_pcf8583(u8 addr, u8 reg, u8 data)
{
	int ret;
	u8 buf[2];
	int msgcnt;
	struct i2c_msg msg[1];

	if (reg & ~ENV_LIB_8583_REG_MSK) {
		return -EINVAL;
	}
		
	buf[0] = reg;
	buf[1] = data;

	msg[0].addr = (addr>>1);
	msg[0].flags = 0;
	msg[0].len = sizeof(buf);
	msg[0].buf = buf;

	msgcnt = sizeof(msg) / sizeof(struct i2c_msg);

	if (!envctrl_adap) {
		BUG();
	}
	ret = i2c_transfer(envctrl_adap, msg, msgcnt);
	if (ret == msgcnt) {
		ret = 0;
	}
	return ret;
}

int envctrl_i2c_write_pcf8583(u8 addr, u8 reg, u8 data)
{
	u8 ret;
	unsigned long flags;

	spin_lock_irqsave(envctrl_lock, flags);
	ret = __envctrl_i2c_write_pcf8583(addr, reg, data);
	spin_unlock_irqrestore(envctrl_lock, flags);
	return ret;
}

