 /* i2c-envctrl - PCF8584 I2C bus nexus driver for supported 
  *               Sun Microsystems systems
  * 
  * Copyright (c) 2004-2005 Eric Brower <ebrower@gmail.com>
  *
  * Adapted from i2c-elektor.c - reference that file for copyright information
  */
 
 #include <linux/kernel.h>
 #include <linux/ioport.h>
 #include <linux/module.h>
 #include <linux/delay.h>
 #include <linux/slab.h>
 #include <linux/init.h>
 #include <linux/interrupt.h>
 #include <linux/pci.h>
 #include <linux/wait.h>
 
 #include <linux/i2c.h>
 #include <linux/i2c-algo-pcf.h>

 #include <asm/of_device.h>
 #include <asm/io.h>
 #include <asm/irq.h>
 #include <asm/i2c-envctrl.h>

 
 #include "asm/i2c-algo-pcf.h"
// #include <linux/i2c-pcf8584.h>

 
 struct pcf8584_regs {
 	volatile u8 data;
 	volatile u8 csr;
 };
 
 static struct __iomem pcf8584_regs *i2c;
 static int irq;
 static int clock  = 0x1c;
 static int own    = 0x55;
 
 static wait_queue_head_t pcf_wait;
 static int pcf_pending;
 static spinlock_t lock;
 
 static struct pcf_envctrl_nexus nexus = { 
 	.type   = TYPE_UNKNOWN, 
 	.i2c_id = -1,
	.ofdev  = NULL,
 	.lock   = SPIN_LOCK_UNLOCKED,
 };
 
 /* ----- local functions ----------------------------------------------	*/
 
 static void pcf_envctrl_setbyte(void *data, int ctl, int val)
 {
 	/* enable irq if any specified for serial operation */
 	if (ctl && irq && (val & I2C_PCF_ESO)) {
 		val |= I2C_PCF_ENI;
 	}
 
 	pr_debug("i2c-envctrl: Write %p 0x%02X\n",
                ctl ? &i2c->csr : &i2c->data, val & 255);
 
 	if (ctl) {
 		writeb(val, &i2c->csr);
 	}
 	else {
 		writeb(val, &i2c->data);
 	}
 }
 
 static int pcf_envctrl_getbyte(void *data, int ctl)
 {
 	int val;
 
 	if (ctl) {
 		val = readb(&i2c->csr);
 	}
 	else {
 		val = readb(&i2c->data);
 	}
 
 	pr_debug("i2c-envctrl: Read %p 0x%02X\n",
 	        ctl ? &i2c->csr : &i2c->data, val);
 
 	return (val);
 }
 
 static int pcf_envctrl_getown(void *data)
 {
 	return (own);
 }
 
 
 static int pcf_envctrl_getclock(void *data)
 {
 	return (clock);
 }
 
 static void pcf_envctrl_waitforpin(void *data) {
 
 	int timeout = 2;
 	long flags;
 
 	if (irq > 0) {
 		spin_lock_irqsave(&lock, flags);
 		if (pcf_pending == 0) {
 			spin_unlock_irqrestore(&lock, flags);
 			if (interruptible_sleep_on_timeout(&pcf_wait,
 								timeout*HZ)) {
 				spin_lock_irqsave(&lock, flags);
 				if (pcf_pending == 1) {
 					pcf_pending = 0;
 				}
 				spin_unlock_irqrestore(&lock, flags);
 			}
 		} else {
 			pcf_pending = 0;
 			spin_unlock_irqrestore(&lock, flags);
 		}
 	} else {
 		udelay(100);
 	}
 }
 
 
 static irqreturn_t pcf_envctrl_handler(int this_irq, void *dev_id) {
 	spin_lock(&lock);
 	pcf_pending = 1;
 	spin_unlock(&lock);
 	wake_up_interruptible(&pcf_wait);
 	return IRQ_HANDLED;
 }
 
 
 static int pcf_envctrl_init(void)
 {
 	spin_lock_init(&lock);
 	if (irq > 0) {
 		if (request_irq(irq, pcf_envctrl_handler, 0, "PCF8584", NULL) < 0) {
 			printk(KERN_ERR "i2c-envctrl: Request irq%d failed\n", irq);
 			irq = 0;
 		} 
 		else {
 			enable_irq(irq);
 		}
 	}
 	return 0;
 }
 
 static struct i2c_algo_pcf_data pcf_envctrl_data = {
 	.setpcf	    = pcf_envctrl_setbyte,
 	.getpcf	    = pcf_envctrl_getbyte,
 	.getown	    = pcf_envctrl_getown,
 	.getclock   = pcf_envctrl_getclock,
 	.waitforpin = pcf_envctrl_waitforpin,
 	.lab_mdelay = 10,
 };

 static struct i2c_adapter pcf_envctrl_ops = {
 	.owner		= THIS_MODULE,
 	.algo_data	= &pcf_envctrl_data,
 	.name		= "PCF8584 bus nexus adapter",
 };


 static int __devinit envctrl_probe(struct of_device *op, const struct of_device_id *match)
{
 	if (!strcmp(op->node->name, "SUNW,envctrltwo")) {
 		nexus.type = TYPE_ENVCTRLTWO;
 	}
 	else if (!strcmp(op->node->name, "SUNW,envctrl")) {
 		nexus.type = TYPE_ENVCTRL;
 	}
	else {
		return -ENODEV;
	}

	nexus.ofdev = op;

 	i2c = of_ioremap(&nexus.ofdev->resource[0], 0, 
			 sizeof(struct pcf8584_regs), "PCF8584");
 	if (!i2c) {
 		return -EINVAL;
 	}
 	init_waitqueue_head(&pcf_wait);
 
 	if (pcf_envctrl_init()) {
 		return -ENODEV;
	}
	
	pcf_envctrl_ops.dev.parent=&op->dev;
 
 	if ((nexus.i2c_id = i2c_pcf_add_bus(&pcf_envctrl_ops)) < 0)
 		goto fail;
 
 	printk(KERN_INFO "i2c-envctrl: %s PCF8584 bus nexus at 0x%lx\n",
 		op->node->name, (unsigned long)i2c);
 	
 	return 0;
 
  fail:
 	if (irq > 0) {
 		disable_irq(irq);
 		free_irq(irq, NULL);
 	}
 	of_iounmap(&nexus.ofdev->resource[0], i2c, 
		   sizeof(struct pcf8584_regs));
 
 	return -ENODEV;
}
 
static struct of_device_id envctrl_match[] = {
	{
		.name = "SUNW,envctrl",		/* E450 */
	},
	{	
		.name = "SUNW,envctrltwo",	/* E250 */
	},
	{},
};

static int __devexit envctrl_remove(struct of_device *dev) 
{
 	i2c_del_adapter(&pcf_envctrl_ops);
 
 	if (irq > 0) {
 		disable_irq(irq);
 		free_irq(irq, NULL);
 	}
 
 	if (i2c) {
 		of_iounmap(&nexus.ofdev->resource[0], i2c, 
			   sizeof(struct pcf8584_regs));
 	}
	return 0;
}

static struct of_platform_driver envctrl_driver = {
	.name 		= "i2c-envctrl",
	.match_table 	= envctrl_match,
	.probe 		= envctrl_probe,
	.remove 	= __devexit_p(envctrl_remove),
};

static int __init i2c_envctrl_init(void) 
{
	return of_register_driver(&envctrl_driver, &of_bus_type);
}
 
 static void i2c_envctrl_exit(void)
 {
	of_unregister_driver(&envctrl_driver);
 }
 
 struct pcf_envctrl_nexus * pcf_envctrl_get_nexus(void)
 {
 	return &nexus;
 }
 
 EXPORT_SYMBOL(pcf_envctrl_get_nexus);
 
 MODULE_AUTHOR("Eric Brower <ebrower@gmail.com>");
 MODULE_DESCRIPTION("PCF8584 I2C bus nexus driver for Sun Microsystems envctrl");
 MODULE_LICENSE("GPL");
 
 module_param(irq, int, 0);
 module_param(clock, int, 0);
 module_param(own, int, 0);
 
 subsys_initcall(i2c_envctrl_init);
 module_exit(i2c_envctrl_exit);
